# Copyright (c) 2025 Cloudification GmbH.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Qos Type interface.
"""

from manilaclient import api_versions
from manilaclient import base
from manilaclient.common import constants


class QosType(base.Resource):
    """A Qos Type represents Quality of service of Manila resource."""

    def __init__(self, manager, info, loaded=False):
        super().__init__(manager, info, loaded)
        self._specs = info.get('specs', {})

    def __repr__(self):
        return f"<QosType: {self.name}>"

    def get_keys(self, prefer_resource_data=True):
        """Get specs from a qos type.

        :param prefer_resource_data: By default specs are retrieved from
        resource data, but user can force this method to make API call.
        :return: dict with specs
        """
        specs = getattr(self, 'specs', None)

        if prefer_resource_data and specs:
            return specs

        qos_type_id = base.getid(self)
        _resp, body = self.manager.api.client.get(
            f"/qos-types/{qos_type_id}/specs"
        )

        self.specs = body["specs"]

        return body["specs"]

    def set_keys(self, metadata):
        """Set specs on a qos type.

        :param metadata: A dict of key/value pairs to be set
        """
        body = {'specs': metadata}
        qos_type_id = base.getid(self)
        return self.manager._create(
            f"/qos-types/{qos_type_id}/specs",
            body,
            "specs",
            return_raw=True,
        )

    def unset_keys(self, keys):
        """Unset specs on a qos type.

        :param keys: A list of keys to be unset
        """
        qos_type_id = base.getid(self)
        for k in keys:
            self.manager._delete(f"/qos-types/{qos_type_id}/specs/{k}")

    def update(self, **kwargs):
        """Update this qos type."""
        return self.manager.update(self, **kwargs)

    def delete(self):
        """Delete this qos type."""
        return self.manager.delete(self)


class QosTypeManager(base.ManagerWithFind):
    """Manage :class:`QosType` resources."""

    resource_class = QosType

    @api_versions.wraps(constants.QOS_TYPE_VERSION)
    def list(self, search_opts=None, sort_key=None, sort_dir=None):
        """Get a list of all qos types.

        :param search_opts: Search options to filter out qos types.
        :param sort_key: Key to be sorted.
        :param sort_dir: Sort direction, should be 'desc' or 'asc'.
        :rtype: list of :class:`QosType`.
        """
        search_opts = search_opts or {}

        if sort_key is not None:
            if sort_key in constants.QOS_TYPE_SORT_KEY_VALUES:
                search_opts['sort_key'] = sort_key
            else:
                raise ValueError(
                    'sort_key must be one of the following: {}.'.format(
                        ', '.join(constants.QOS_TYPE_SORT_KEY_VALUES)
                    )
                )

        if sort_dir is not None:
            if sort_dir in constants.SORT_DIR_VALUES:
                search_opts['sort_dir'] = sort_dir
            else:
                raise ValueError(
                    'sort_dir must be one of the following: {}.'.format(
                        ', '.join(constants.SORT_DIR_VALUES)
                    )
                )

        query_string = self._build_query_string(search_opts)
        return self._list(f"/qos-types{query_string}", "qos_types")

    @api_versions.wraps(constants.QOS_TYPE_VERSION)
    def get(self, qos_type):
        """Get a specific qos type.

        :param qos_type: The ID of the :class:`QosType` to get.
        :rtype: :class:`QosType`
        """
        qos_type_id = base.getid(qos_type)
        return self._get(f"/qos-types/{qos_type_id}", "qos_type")

    @api_versions.wraps(constants.QOS_TYPE_VERSION)
    def delete(self, qos_type):
        """Delete a specific qos_type.

        :param qos_type: The name or ID of the :class:`QosType` to get.
        """
        qos_type_id = base.getid(qos_type)
        self._delete(f"/qos-types/{qos_type_id}")

    @api_versions.wraps(constants.QOS_TYPE_VERSION)
    def create(self, name, description=None, specs=None):
        """Create a qos type.

        :param name: Descriptive name of the qos type
        :param specs: Specs of the qos type
        :rtype: :class:`QosType`
        """
        if specs is None:
            specs = {}

        body = {
            "qos_type": {
                "name": name,
                "specs": specs,
            }
        }

        if description:
            body["qos_type"]["description"] = description
        return self._create("/qos-types", body, "qos_type")

    @api_versions.wraps(constants.QOS_TYPE_VERSION)
    def update(self, qos_type, **kwargs):
        """Update the description for a qos type.

        :param qos_type: the ID of the :class: `QosType` to update.
        :param description: Description of the qos type.
        :rtype: :class:`QosType`
        """

        if not kwargs:
            return

        body = {
            'qos_type': kwargs,
        }
        qos_type_id = base.getid(qos_type)
        return self._update(f"/qos-types/{qos_type_id}", body, "qos_type")



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging

from osc_lib.command import command
from osc_lib import exceptions
from osc_lib import utils as osc_utils

from manilaclient.common._i18n import _
from manilaclient.common.apiclient import utils as apiutils
from manilaclient.common import constants


LOG = logging.getLogger(__name__)

TRANSFER_DETAIL_ATTRIBUTES = [
    'id',
    'created_at',
    'name',
    'resource_type',
    'resource_id',
    'source_project_id',
    'destination_project_id',
    'accepted',
    'expires_at',
]

TRANSFER_SUMMARY_ATTRIBUTES = [
    'id',
    'name',
    'resource_type',
    'resource_id',
]


class CreateShareTransfer(command.ShowOne):
    """Create a new share transfer."""

    _description = _("Create a new share transfer")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'share', metavar='<share>', help='Name or ID of share to transfer.'
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            default=None,
            help='Transfer name. Default=None.',
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share
        share = osc_utils.find_resource(share_client.shares, parsed_args.share)
        transfer = share_client.transfers.create(
            share.id, name=parsed_args.name
        )

        transfer._info.pop('links', None)

        return self.dict2columns(transfer._info)


class DeleteShareTransfer(command.Command):
    """Remove one or more transfers."""

    _description = _("Remove one or more transfers")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'transfer',
            metavar='<transfer>',
            nargs='+',
            help='Name(s) or ID(s) of the transfer(s).',
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share
        failure_count = 0

        for transfer in parsed_args.transfer:
            try:
                transfer_obj = apiutils.find_resource(
                    share_client.transfers, transfer
                )
                share_client.transfers.delete(transfer_obj.id)
            except Exception as e:
                failure_count += 1
                LOG.error(
                    _("Failed to delete %(transfer)s: %(e)s"),
                    {'transfer': transfer, 'e': e},
                )

        if failure_count > 0:
            raise exceptions.CommandError(
                _("Unable to delete some or all of the specified transfers.")
            )


class ListShareTransfer(command.Lister):
    """Lists all transfers."""

    _description = _("Lists all transfers")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--all-projects',
            action='store_true',
            help=_("Shows details for all tenants. (Admin only)."),
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            default=None,
            help='Filter share transfers by name. Default=None.',
        )
        parser.add_argument(
            '--id',
            metavar='<id>',
            default=None,
            help='Filter share transfers by ID. Default=None.',
        )
        parser.add_argument(
            '--resource-type',
            '--resource_type',
            metavar='<resource_type>',
            default=None,
            help='Filter share transfers by resource type, '
            'which can be share. Default=None.',
        )
        parser.add_argument(
            '--resource-id',
            '--resource_id',
            metavar='<resource_id>',
            default=None,
            help='Filter share transfers by resource ID. Default=None.',
        )
        parser.add_argument(
            '--source-project-id',
            '--source_project_id',
            metavar='<source_project_id>',
            default=None,
            help='Filter share transfers by ID of the Project that '
            'initiated the transfer. Default=None.',
        )
        parser.add_argument(
            '--limit',
            metavar='<limit>',
            type=int,
            default=None,
            help='Maximum number of transfer records to '
            'return. (Default=None)',
        )
        parser.add_argument(
            '--offset',
            metavar="<offset>",
            default=None,
            help='Start position of transfer records listing.',
        )
        parser.add_argument(
            '--sort-key',
            '--sort_key',
            metavar='<sort_key>',
            type=str,
            default=None,
            help=(
                f'Key to be sorted, available keys are '
                f'{constants.SHARE_TRANSFER_SORT_KEY_VALUES}. '
                f'Default=None.'
            ),
        )
        parser.add_argument(
            '--sort-dir',
            '--sort_dir',
            metavar='<sort_dir>',
            type=str,
            default=None,
            help=(
                f'Sort direction, available values are '
                f'{constants.SORT_DIR_VALUES}. '
                f'OPTIONAL: Default=None.'
            ),
        )
        parser.add_argument(
            '--detailed',
            dest='detailed',
            metavar='<0|1>',
            nargs='?',
            type=int,
            const=1,
            default=0,
            help="Show detailed information about filtered share transfers.",
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        columns = ['ID', 'Name', 'Resource Type', 'Resource Id']

        if parsed_args.detailed:
            columns.extend(
                [
                    'Created At',
                    'Source Project Id',
                    'Destination Project Id',
                    'Accepted',
                    'Expires At',
                ]
            )

        search_opts = {
            'all_tenants': parsed_args.all_projects,
            'id': parsed_args.id,
            'name': parsed_args.name,
            'limit': parsed_args.limit,
            'offset': parsed_args.offset,
            'resource_type': parsed_args.resource_type,
            'resource_id': parsed_args.resource_id,
            'source_project_id': parsed_args.source_project_id,
        }

        transfers = share_client.transfers.list(
            detailed=parsed_args.detailed,
            search_opts=search_opts,
            sort_key=parsed_args.sort_key,
            sort_dir=parsed_args.sort_dir,
        )

        return (
            columns,
            (osc_utils.get_item_properties(m, columns) for m in transfers),
        )


class ShowShareTransfer(command.ShowOne):
    """Show details about a share transfer."""

    _description = _("Show details about a share transfer")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'transfer',
            metavar='<transfer>',
            help=_('Name or ID of transfer to show.'),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        transfer = apiutils.find_resource(
            share_client.transfers, parsed_args.transfer
        )

        return (
            TRANSFER_DETAIL_ATTRIBUTES,
            osc_utils.get_dict_properties(
                transfer._info, TRANSFER_DETAIL_ATTRIBUTES
            ),
        )


class AcceptShareTransfer(command.Command):
    """Accepts a share transfer."""

    _description = _("Accepts a share transfer")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'transfer', metavar='<transfer>', help='ID of transfer to accept.'
        )
        parser.add_argument(
            'auth_key',
            metavar='<auth_key>',
            help='Authentication key of transfer to accept.',
        )
        parser.add_argument(
            '--clear-rules',
            '--clear_rules',
            dest='clear_rules',
            action='store_true',
            default=False,
            help="Whether manila should clean up the access rules after the "
            "transfer is complete. (Default=False)",
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        share_client.transfers.accept(
            parsed_args.transfer,
            parsed_args.auth_key,
            clear_access_rules=parsed_args.clear_rules,
        )



#   Copyright 2021 Red Hat Inc. All rights reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from osc_lib import utils as osc_lib_utils

from manilaclient.osc.v2 import share_snapshot_instance_export_locations
from manilaclient.tests.unit.osc import osc_utils
from manilaclient.tests.unit.osc.v2 import fakes as manila_fakes

COLUMNS = ['ID', 'Path', 'Is Admin only']


class TestShareSnapshotInstanceExportLocation(manila_fakes.TestShare):
    def setUp(self):
        super().setUp()

        self.share_snapshot_instances_mock = (
            self.app.client_manager.share.share_snapshot_instances
        )
        self.share_snapshot_instances_mock.reset_mock()

        self.share_snapshot_instances_el_mock = (
            self.app.client_manager.share.share_snapshot_instance_export_locations  # noqa: E501
        )
        self.share_snapshot_instances_el_mock.reset_mock()


class TestShareSnapshotInstanceExportLocationList(
    TestShareSnapshotInstanceExportLocation
):
    def setUp(self):
        super().setUp()

        self.share_snapshot_instance = manila_fakes.FakeShareSnapshotIntances.create_one_snapshot_instance()  # noqa: E501

        self.share_snapshot_instances_export_locations = manila_fakes.FakeShareSnapshotInstancesExportLocations.create_share_snapshot_instances(  # noqa: E501
            count=2
        )

        self.share_snapshot_instances_mock.get.return_value = (
            self.share_snapshot_instance
        )

        self.share_snapshot_instances_el_mock.list.return_value = (
            self.share_snapshot_instances_export_locations
        )

        self.cmd = share_snapshot_instance_export_locations.ShareSnapshotInstanceExportLocationList(  # noqa: E501
            self.app, None
        )

    def test_share_snapshot_instance_export_location_list_missing_args(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            osc_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_share_snapshot_instance_export_location_list(self):
        values = (
            osc_lib_utils.get_dict_properties(s._info, COLUMNS)
            for s in self.share_snapshot_instances_export_locations
        )
        arglist = [self.share_snapshot_instance.id]
        verifylist = [('instance', self.share_snapshot_instance.id)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.assertEqual(COLUMNS, columns)
        self.assertEqual(list(values), list(data))


class TestShareSnapshotInstanceExportLocationShow(
    TestShareSnapshotInstanceExportLocation
):
    def setUp(self):
        super().setUp()

        self.share_snapshot_instance = manila_fakes.FakeShareSnapshotIntances.create_one_snapshot_instance()  # noqa: E501

        self.share_snapshot_instances_export_location = (
            manila_fakes.FakeShareSnapshotInstancesExportLocations.create_one_snapshot_instance()  # noqa: E501
        )

        self.share_snapshot_instances_mock.get.return_value = (
            self.share_snapshot_instance
        )

        self.share_snapshot_instances_el_mock.get.return_value = (
            self.share_snapshot_instances_export_location
        )

        self.cmd = share_snapshot_instance_export_locations.ShareSnapshotInstanceExportLocationShow(  # noqa: E501
            self.app, None
        )

        self.data = (
            self.share_snapshot_instances_export_location._info.values()
        )
        self.columns = (
            self.share_snapshot_instances_export_location._info.keys()
        )

    def test_share_snapshot_instance_export_location_show_missing_args(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            osc_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_share_snapshot_instance_export_location_show(self):
        arglist = [
            self.share_snapshot_instance.id,
            self.share_snapshot_instances_export_location.id,
        ]
        verifylist = [
            ('snapshot_instance', self.share_snapshot_instance.id),
            (
                'export_location',
                self.share_snapshot_instances_export_location.id,
            ),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.share_snapshot_instances_mock.get.assert_called_with(
            self.share_snapshot_instance.id
        )
        self.share_snapshot_instances_el_mock.get.assert_called_with(
            self.share_snapshot_instances_export_location.id,
            snapshot_instance=self.share_snapshot_instance,
        )
        self.assertCountEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from tempest.lib import exceptions

from manilaclient.tests.functional.osc import base


class ShareInstancesCLITest(base.OSCClientTestBase):
    def setUp(self):
        super().setUp()
        # Create a share to test with
        self.share = self.create_share()
        # Get share instances for testing
        self.share_instances = self.listing_result('share instance', 'list')
        self.share_instance = (
            self.share_instances[0] if self.share_instances else None
        )
        self.non_existent_id = '76a7dd8e-7564-41d4-a184-06dc20e13d7a'
        self.invalid_id = 'invalid-format'

    def test_share_instance_list(self):
        # Test basic listing
        instances = self.listing_result('share instance', 'list')
        self.assertTableStruct(
            instances,
            [
                'ID',
                'Share ID',
                'Host',
                'Status',
                'Availability Zone',
                'Share Network ID',
                'Share Server ID',
                'Share Type ID',
            ],
        )

    def test_share_instance_show(self):
        result = self.dict_result(
            'share instance', f'show {self.share_instance["ID"]}'
        )
        self.assertEqual(self.share_instance['ID'], result['id'])

    def test_share_instance_delete_non_existent(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.openstack,
            f'share instance delete {self.non_existent_id}',
        )

    def test_share_instance_set_non_existent(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.openstack,
            f'share instance set {self.non_existent_id}  --status available',
        )

    def test_share_instance_set_invalid_status(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.openstack,
            f'share instance set {self.invalid_id} --status available',
        )

    def test_share_instance_show_non_existent(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.openstack,
            f'share instance show {self.non_existent_id}',
        )

    def test_share_instance_show_invalid_id_format(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.openstack,
            f'share instance show {self.invalid_id}',
        )

    def test_share_instance_list_invalid_filter_share_id_format(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.openstack,
            f'share instance list --share {self.invalid_id}',
        )

    def test_share_instance_list_invalid_filter_non_existent(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.openstack,
            f'share instance list --share {self.non_existent_id}',
        )



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import json

from manilaclient.tests.functional.osc import base


class ShareBackupCLITest(base.OSCClientTestBase):
    """Functional tests for share backup."""

    def test_share_backup_create(self):
        share = self.create_share()

        backup = self.create_backup(
            share_id=share['id'],
            name='test_backup_create',
            description='Description',
            backup_options={'dummy': True},
        )

        # fetch latest after periodic callback updates status
        backup = json.loads(
            self.openstack(f'share backup show -f json {backup["id"]}')
        )

        self.assertEqual(share["id"], backup["share_id"])
        self.assertEqual('test_backup_create', backup["name"])
        self.assertEqual('Description', backup["description"])
        self.assertEqual('available', backup["status"])

        backups_list = self.listing_result('share backup', 'list')
        self.assertIn(backup['id'], [item['ID'] for item in backups_list])

    def test_share_backup_delete(self):
        share = self.create_share()

        backup = self.create_backup(
            share_id=share['id'],
            backup_options={'dummy': True},
            add_cleanup=False,
        )

        self.openstack(f'share backup delete {backup["id"]} --wait')

        self.check_object_deleted('share backup', backup["id"])

    def test_share_backup_show(self):
        share = self.create_share()

        backup = self.create_backup(
            share_id=share['id'],
            name='test_backup_show',
            description='Description',
            backup_options={'dummy': True},
        )

        show_result = self.dict_result('share backup', f'show {backup["id"]}')

        self.assertEqual(backup["id"], show_result["id"])
        self.assertEqual('test_backup_show', show_result["name"])
        self.assertEqual('Description', show_result["description"])

    def test_share_backup_restore(self):
        share = self.create_share()

        backup = self.create_backup(
            share_id=share['id'], backup_options={'dummy': True}
        )

        self.openstack(f'share backup restore {backup["id"]} --wait')

        backup = json.loads(
            self.openstack(f'share backup show -f json {backup["id"]}')
        )
        share = json.loads(self.openstack(f'share show -f json {share["id"]}'))

        self.assertEqual('available', backup['status'])
        self.assertEqual('available', share['status'])

    def test_share_backup_set(self):
        share = self.create_share()

        backup = self.create_backup(
            share_id=share['id'], backup_options={'dummy': True}
        )

        self.openstack(
            f'share backup set {backup["id"]} '
            f'--name test_backup_set --description Description'
        )

        show_result = self.dict_result('share backup ', f'show {backup["id"]}')

        self.assertEqual(backup['id'], show_result["id"])
        self.assertEqual('test_backup_set', show_result["name"])
        self.assertEqual('Description', show_result["description"])

    def test_share_backup_unset(self):
        share = self.create_share()

        backup = self.create_backup(
            share_id=share['id'],
            name='test_backup_unset',
            description='Description',
            backup_options={'dummy': True},
        )

        self.openstack(
            f'share backup unset {backup["id"]} --name --description'
        )

        show_result = json.loads(
            self.openstack(f'share backup show -f json {backup["id"]}')
        )

        self.assertEqual(backup['id'], show_result["id"])
        self.assertIsNone(show_result["name"])
        self.assertIsNone(show_result["description"])

    def test_share_backup_list(self):
        share_1 = self.create_share()
        share_2 = self.create_share()

        backup_1 = self.create_backup(
            share_id=share_1['id'], backup_options={'dummy': True}
        )
        backup_2 = self.create_backup(
            share_id=share_2['id'], backup_options={'dummy': True}
        )

        backups_list = self.listing_result(
            'share backup', f'list --name {backup_2["name"]} '
        )

        self.assertTableStruct(
            backups_list, ['ID', 'Name', 'Share ID', 'Status']
        )
        self.assertEqual(1, len(backups_list))
        self.assertIn(backup_2['id'], [item['ID'] for item in backups_list])

        backups_list = self.listing_result(
            'share backup', f'list --share {share_1["id"]} --detail'
        )

        self.assertTableStruct(
            backups_list,
            [
                'ID',
                'Name',
                'Share ID',
                'Status',
                'Description',
                'Availability Zone',
                'Created At',
                'Updated At',
                'Size',
                'Progress',
                'Restore Progress',
                'Host',
                'Topic',
            ],
        )
        self.assertEqual(1, len(backups_list))
        self.assertIn(backup_1['id'], [item['ID'] for item in backups_list])



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging

from openstackclient.identity import common as identity_common
from osc_lib.command import command
from osc_lib import exceptions
from osc_lib import utils as osc_utils
from oslo_utils import uuidutils

from manilaclient.common._i18n import _
from manilaclient.common.apiclient import utils as apiutils
from manilaclient.common import constants


LOG = logging.getLogger(__name__)


LOCK_DETAIL_ATTRIBUTES = [
    'ID',
    'Resource Id',
    'Resource Type',
    'Resource Action',
    'Lock Context',
    'User Id',
    'Project Id',
    'Created At',
    'Updated At',
    'Lock Reason',
]

LOCK_SUMMARY_ATTRIBUTES = [
    'ID',
    'Resource Id',
    'Resource Type',
    'Resource Action',
]

RESOURCE_TYPE_MANAGERS = {
    'share': 'shares',
    'access_rule': 'share_access_rules',
}


class CreateResourceLock(command.ShowOne):
    """Create a new resource lock."""

    _description = _("Lock a resource action from occurring on a resource")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'resource',
            metavar='<resource_name_or_id>',
            help='Name or ID of resource to lock.',
        )
        parser.add_argument(
            'resource_type',
            metavar='<resource_type>',
            help='Type of the resource (e.g.: share, access).',
        )
        parser.add_argument(
            '--resource-action',
            '--resource_action',
            metavar='<resource_action>',
            default='delete',
            help='Action to lock on the resource (default="delete")',
        )
        parser.add_argument(
            '--lock-reason',
            '--lock_reason',
            '--reason',
            metavar='<lock_reason>',
            help='Reason for the resource lock.',
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share
        resource_type = parsed_args.resource_type
        if resource_type not in RESOURCE_TYPE_MANAGERS:
            raise exceptions.CommandError(_("Unsupported resource type"))
        res_manager = RESOURCE_TYPE_MANAGERS[resource_type]

        resource = osc_utils.find_resource(
            getattr(share_client, res_manager), parsed_args.resource
        )
        resource_lock = share_client.resource_locks.create(
            resource.id,
            resource_type,
            parsed_args.resource_action,
            parsed_args.lock_reason,
        )

        resource_lock._info.pop('links', None)

        return self.dict2columns(resource_lock._info)


class DeleteResourceLock(command.Command):
    """Remove one or more resource locks."""

    _description = _("Remove one or more resource locks")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'lock', metavar='<lock>', nargs='+', help='ID(s) of the lock(s).'
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share
        failure_count = 0

        for lock in parsed_args.lock:
            try:
                lock = apiutils.find_resource(
                    share_client.resource_locks, lock
                )
                lock.delete()
            except Exception as e:
                failure_count += 1
                LOG.error(
                    _("Failed to delete %(lock)s: %(e)s"),
                    {'lock': lock, 'e': e},
                )

        if failure_count > 0:
            raise exceptions.CommandError(
                _("Unable to delete some or all of the specified locks.")
            )


class ListResourceLock(command.Lister):
    """Lists all resource locks."""

    _description = _("Lists all resource locks")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--all-projects',
            action='store_true',
            help=_("Filter resource locks for all projects. (Admin only)."),
        )
        parser.add_argument(
            '--project',
            default=None,
            help=_(
                "Filter resource locks for specific project by name or ID, "
                "combine with --all-projects (Admin only)."
            ),
        )
        parser.add_argument(
            '--user',
            default=None,
            help=_(
                "Filter resource locks for specific user by name or ID, "
                "combine with --all-projects to search across projects "
                "(Admin only)."
            ),
        )
        parser.add_argument(
            '--id',
            metavar='<id>',
            default=None,
            help='Filter resource locks by ID. Default=None.',
        )
        parser.add_argument(
            '--resource',
            '--resource-id',
            '--resource_id',
            default=None,
            metavar='<resource-id>',
            dest='resource',
            help=_(
                "Filter resource locks for a resource by ID, specify "
                "--resource-type to look up by name."
            ),
        )
        parser.add_argument(
            '--resource-type',
            '--resource_type',
            default=None,
            metavar='<resource_type>',
            help=_("Filter resource locks by type of resource."),
        )
        parser.add_argument(
            '--resource-action',
            '--resource_action',
            default=None,
            metavar='<resource_action>',
            help=_("Filter resource locks by resource action."),
        )

        parser.add_argument(
            '--lock-context',
            '--lock_context',
            '--context',
            default=None,
            choices=['user', 'admin', 'service'],
            metavar='<lock_context>',
            help=_("Filter resource locks by context."),
        )
        parser.add_argument(
            '--since',
            default=None,
            metavar='<created_since>',
            help=_(
                "Filter resource locks created since given date. "
                "The date format must be conforming to ISO8601. "
            ),
        )
        parser.add_argument(
            '--before',
            default=None,
            metavar='<created_before>',
            help=_(
                "Filter resource locks created before given date. "
                "The date format must be conforming to ISO8601. "
            ),
        )
        parser.add_argument(
            '--limit',
            metavar='<limit>',
            type=int,
            default=None,
            help=_("Number of resource locks to list. (Default=None)"),
        )
        parser.add_argument(
            '--offset',
            metavar="<offset>",
            default=None,
            help='Starting position of resource lock records '
            'in a paginated list.',
        )
        parser.add_argument(
            '--sort-key',
            '--sort_key',
            metavar='<sort_key>',
            type=str,
            default=None,
            choices=constants.RESOURCE_LOCK_SORT_KEY_VALUES,
            help=(
                f'Key to be sorted, available keys are '
                f'{constants.RESOURCE_LOCK_SORT_KEY_VALUES}. Default=None.'
            ),
        )
        parser.add_argument(
            '--sort-dir',
            '--sort_dir',
            metavar='<sort_dir>',
            type=str,
            default=None,
            choices=constants.SORT_DIR_VALUES,
            help=(
                f'Sort direction, available values are '
                f'{constants.SORT_DIR_VALUES}. OPTIONAL: Default=None.'
            ),
        )
        parser.add_argument(
            '--detailed',
            dest='detailed',
            metavar='<0|1>',
            nargs='?',
            type=int,
            const=1,
            default=0,
            help="Show detailed information about filtered resource locks.",
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share
        identity_client = self.app.client_manager.identity

        columns = (
            LOCK_SUMMARY_ATTRIBUTES
            if not parsed_args.detailed
            else LOCK_DETAIL_ATTRIBUTES
        )

        project_id = None
        user_id = None

        if parsed_args.project:
            project_id = identity_common.find_project(
                identity_client,
                parsed_args.project,
                parsed_args.project_domain,
            ).id
        if parsed_args.user:
            user_id = identity_common.find_user(
                identity_client, parsed_args.user, parsed_args.user_domain
            ).id
        # set all_projects when using project option
        all_projects = bool(parsed_args.project) or parsed_args.all_projects

        resource_id = parsed_args.resource
        resource_type = parsed_args.resource_type
        if resource_type is not None:
            if resource_type not in RESOURCE_TYPE_MANAGERS:
                raise exceptions.CommandError(_("Unsupported resource type"))
            if resource_id is not None:
                res_manager = RESOURCE_TYPE_MANAGERS[resource_type]
                resource_id = osc_utils.find_resource(
                    getattr(share_client, res_manager), parsed_args.resource
                ).id
        elif resource_id and not uuidutils.is_uuid_like(resource_id):
            raise exceptions.CommandError(
                _("Provide resource ID or specify --resource-type.")
            )

        search_opts = {
            'all_projects': all_projects,
            'project_id': project_id,
            'user_id': user_id,
            'id': parsed_args.id,
            'resource_id': resource_id,
            'resource_type': parsed_args.resource_type,
            'resource_action': parsed_args.resource_action,
            'lock_context': parsed_args.lock_context,
            'created_before': parsed_args.before,
            'created_since': parsed_args.since,
            'limit': parsed_args.limit,
            'offset': parsed_args.offset,
        }

        resource_locks = share_client.resource_locks.list(
            search_opts=search_opts,
            sort_key=parsed_args.sort_key,
            sort_dir=parsed_args.sort_dir,
        )

        return (
            columns,
            (
                osc_utils.get_item_properties(m, columns)
                for m in resource_locks
            ),
        )


class ShowResourceLock(command.ShowOne):
    """Show details about a resource lock."""

    _description = _("Show details about a resource lock")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'lock', metavar='<lock>', help=_('ID of resource lock to show.')
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        resource_lock = apiutils.find_resource(
            share_client.resource_locks, parsed_args.lock
        )

        return (
            LOCK_DETAIL_ATTRIBUTES,
            osc_utils.get_dict_properties(
                resource_lock._info, LOCK_DETAIL_ATTRIBUTES
            ),
        )


class SetResourceLock(command.Command):
    """Set resource lock properties."""

    _description = _("Update resource lock properties")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'lock', metavar='<lock>', help='ID of lock to update.'
        )
        parser.add_argument(
            '--resource-action',
            '--resource_action',
            metavar='<resource_action>',
            help='Resource action to set in the resource lock',
        )
        parser.add_argument(
            '--lock-reason',
            '--lock_reason',
            '--reason',
            dest='lock_reason',
            help="Reason for the resource lock",
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        update_kwargs = {}
        if parsed_args.resource_action is not None:
            update_kwargs['resource_action'] = parsed_args.resource_action
        if parsed_args.lock_reason is not None:
            update_kwargs['lock_reason'] = parsed_args.lock_reason
        if update_kwargs:
            share_client.resource_locks.update(
                parsed_args.lock, **update_kwargs
            )


class UnsetResourceLock(command.Command):
    """Unsets a property on a resource lock."""

    _description = _("Remove resource lock properties")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'lock', metavar='<lock>', help='ID of resource lock to update.'
        )
        parser.add_argument(
            '--lock-reason',
            '--lock_reason',
            '--reason',
            dest='lock_reason',
            action='store_true',
            default=False,
            help="Unset the lock reason. (Default=False)",
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        if parsed_args.lock_reason:
            share_client.resource_locks.update(
                parsed_args.lock, lock_reason=None
            )



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
from operator import xor

from osc_lib.cli import format_columns
from osc_lib.cli import parseractions
from osc_lib.command import command
from osc_lib import exceptions
from osc_lib import utils as oscutils

from manilaclient import api_versions
from manilaclient.common._i18n import _
from manilaclient.common import cliutils


LOG = logging.getLogger(__name__)


class CreateShareNetworkSubnet(command.ShowOne):
    """Create a share network subnet."""

    _description = _("Create a share network subnet")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "share_network",
            metavar="<share-network>",
            help=_("Share network name or ID."),
        )
        parser.add_argument(
            "--neutron-net-id",
            metavar="<neutron-net-id>",
            default=None,
            help=_(
                "Neutron network ID. Used to set up network for share "
                "servers (optional). Should be defined together with "
                "'--neutron-subnet-id'."
            ),
        )
        parser.add_argument(
            "--neutron-subnet-id",
            metavar="<neutron-subnet-id>",
            default=None,
            help=_(
                "Neutron subnet ID. Used to set up network for share "
                "servers (optional). Should be defined together with "
                "'--neutron-net-id' to which this subnet belongs to. "
            ),
        )
        parser.add_argument(
            "--availability-zone",
            metavar="<availability-zone>",
            default=None,
            help=_(
                "Optional availability zone that the subnet is available "
                "within (Default=None). If None, the subnet will be "
                "considered as being available across all availability "
                "zones."
            ),
        )
        parser.add_argument(
            '--check-only',
            default=False,
            action='store_true',
            help=_(
                "Run a dry-run of a share network subnet create. "
                "Available only for microversion >= 2.70."
            ),
        )
        parser.add_argument(
            '--restart-check',
            default=False,
            action='store_true',
            help=_(
                "Restart a dry-run of a share network subnet create. "
                "Helpful when check results are stale. "
                "Available only for microversion >= 2.70."
            ),
        )
        parser.add_argument(
            "--property",
            metavar="<key=value>",
            default={},
            action=parseractions.KeyValueAction,
            help=_(
                "Set a property to this share network subnet "
                "(repeat option to set multiple properties). "
                "Available only for microversion >= 2.78."
            ),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        # check and restart check during create is only available from 2.70.
        if (
            parsed_args.check_only
            and share_client.api_version < api_versions.APIVersion("2.70")
        ):
            raise exceptions.CommandError(
                "Check only can be specified only with manila API "
                "version >= 2.70."
            )
        if (
            parsed_args.restart_check
            and share_client.api_version < api_versions.APIVersion("2.70")
        ):
            raise exceptions.CommandError(
                "Restart check can be specified only with manila API "
                "version >= 2.70."
            )

        if (
            parsed_args.property
            and share_client.api_version < api_versions.APIVersion("2.78")
        ):
            raise exceptions.CommandError(
                "Property can be specified only with manila API "
                "version >= 2.78."
            )

        neutron_client = getattr(self.app.client_manager, 'network', None)
        neutron_net_id = parsed_args.neutron_net_id
        neutron_subnet_id = parsed_args.neutron_subnet_id

        if xor(bool(neutron_net_id), bool(neutron_subnet_id)):
            raise exceptions.CommandError(
                "Both neutron_net_id and neutron_subnet_id should be "
                "specified. Alternatively, neither of them should be "
                "specified."
            )

        if neutron_client and neutron_net_id:
            try:
                neutron_net_id = neutron_client.find_network(
                    neutron_net_id, ignore_missing=False
                ).id
            except Exception:
                raise exceptions.CommandError(
                    f"Neutron network '{neutron_net_id}' not found."
                )

        if neutron_client and neutron_subnet_id:
            try:
                neutron_subnet_id = neutron_client.find_subnet(
                    neutron_subnet_id, ignore_missing=False
                ).id
            except Exception:
                raise exceptions.CommandError(
                    f"Neutron subnet '{neutron_subnet_id}' not found."
                )

        share_network_id = oscutils.find_resource(
            share_client.share_networks, parsed_args.share_network
        ).id

        if parsed_args.check_only or parsed_args.restart_check:
            if parsed_args.property:
                raise exceptions.CommandError(
                    "Property cannot be specified with check operation."
                )

            subnet_create_check = (
                share_client.share_networks.share_network_subnet_create_check(
                    neutron_net_id=neutron_net_id,
                    neutron_subnet_id=neutron_subnet_id,
                    availability_zone=parsed_args.availability_zone,
                    reset_operation=parsed_args.restart_check,
                    share_network_id=share_network_id,
                )
            )
            subnet_data = subnet_create_check[1]
            if subnet_data:
                if parsed_args.formatter == 'table':
                    for k, v in subnet_data.items():
                        if isinstance(v, dict):
                            capabilities_list = [v]
                            dict_values = cliutils.convert_dict_list_to_string(
                                capabilities_list
                            )
                            subnet_data[k] = dict_values
        else:
            share_network_subnet = share_client.share_network_subnets.create(
                neutron_net_id=neutron_net_id,
                neutron_subnet_id=neutron_subnet_id,
                availability_zone=parsed_args.availability_zone,
                share_network_id=share_network_id,
                metadata=parsed_args.property,
            )
            subnet_data = share_network_subnet._info

        return self.dict2columns(subnet_data)


class DeleteShareNetworkSubnet(command.Command):
    """Delete a share network subnet."""

    _description = _("Delete a share network subnet")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "share_network",
            metavar="<share-network>",
            help=_("Share network name or ID."),
        )
        parser.add_argument(
            "share_network_subnet",
            metavar="<share-network-subnet>",
            nargs="+",
            help=_("ID(s) of share network subnet(s) to be deleted."),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share
        result = 0

        share_network_id = oscutils.find_resource(
            share_client.share_networks, parsed_args.share_network
        ).id

        for subnet in parsed_args.share_network_subnet:
            try:
                share_client.share_network_subnets.delete(
                    share_network_id, subnet
                )

            except Exception as e:
                result += 1
                LOG.error(
                    f"Failed to delete share network subnet with "
                    f"ID {subnet}: {e}"
                )
        if result > 0:
            total = len(parsed_args.share_network_subnet)
            raise exceptions.CommandError(
                f"{result} of {total} share network subnets failed to be "
                f"deleted."
            )


class ShowShareNetworkSubnet(command.ShowOne):
    """Show share network subnet."""

    _description = _("Show share network subnet")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "share_network",
            metavar="<share-network>",
            help=_("Share network name or ID."),
        )
        parser.add_argument(
            "share_network_subnet",
            metavar="<share-network-subnet>",
            help=_("ID of share network subnet to show."),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        share_network_id = oscutils.find_resource(
            share_client.share_networks, parsed_args.share_network
        ).id

        share_network_subnet = share_client.share_network_subnets.get(
            share_network_id, parsed_args.share_network_subnet
        )
        data = share_network_subnet._info

        # Special mapping for columns to make the output easier to read:
        # 'metadata' --> 'properties'
        data.update(
            {
                'properties': format_columns.DictColumn(
                    data.pop('metadata', {})
                ),
            },
        )

        return self.dict2columns(data)


class SetShareNetworkSubnet(command.Command):
    """Set share network subnet properties."""

    _description = _("Set share network subnet properties")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "share_network",
            metavar="<share-network>",
            help=_("Share network name or ID."),
        )
        parser.add_argument(
            "share_network_subnet",
            metavar="<share-network-subnet>",
            help=_("ID of share network subnet to set a property."),
        )
        parser.add_argument(
            "--property",
            metavar="<key=value>",
            default={},
            action=parseractions.KeyValueAction,
            help=_(
                "Set a property to this share network subnet "
                "(repeat option to set multiple properties). "
                "Available only for microversion >= 2.78."
            ),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        if (
            parsed_args.property
            and share_client.api_version < api_versions.APIVersion("2.78")
        ):
            raise exceptions.CommandError(
                "Property can be specified only with manila API "
                "version >= 2.78."
            )

        share_network_id = oscutils.find_resource(
            share_client.share_networks, parsed_args.share_network
        ).id

        if parsed_args.property:
            try:
                share_client.share_network_subnets.set_metadata(
                    share_network_id,
                    parsed_args.property,
                    subresource=parsed_args.share_network_subnet,
                )
            except Exception as e:
                raise exceptions.CommandError(
                    _("Failed to set subnet property '%(properties)s': %(e)s")
                    % {'properties': parsed_args.property, 'e': e}
                )


class UnsetShareNetworkSubnet(command.Command):
    """Unset a share network subnet property."""

    _description = _("Unset a share network subnet property")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "share_network",
            metavar="<share-network>",
            help=_("Share network name or ID."),
        )
        parser.add_argument(
            "share_network_subnet",
            metavar="<share-network-subnet>",
            help=_("ID of share network subnet to set a property."),
        )
        parser.add_argument(
            '--property',
            metavar='<key>',
            action='append',
            help=_(
                "Remove a property from share network subnet "
                "(repeat option to remove multiple properties). "
                "Available only for microversion >= 2.78."
            ),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        if (
            parsed_args.property
            and share_client.api_version < api_versions.APIVersion("2.78")
        ):
            raise exceptions.CommandError(
                "Property can be specified only with manila API "
                "version >= 2.78."
            )

        share_network_id = oscutils.find_resource(
            share_client.share_networks, parsed_args.share_network
        ).id

        if parsed_args.property:
            result = 0
            for key in parsed_args.property:
                try:
                    share_client.share_network_subnets.delete_metadata(
                        share_network_id,
                        [key],
                        subresource=parsed_args.share_network_subnet,
                    )
                except Exception as e:
                    result += 1
                    LOG.error(
                        "Failed to unset subnet property '%(key)s': %(e)s",
                        {'key': key, 'e': e},
                    )
            if result > 0:
                total = len(parsed_args.property)
                raise exceptions.CommandError(
                    f"{result} of {total} subnet properties failed to be "
                    f"unset."
                )



#   Copyright 2019 Red Hat Inc. All rights reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#


from unittest import mock
import uuid

import ddt
from osc_lib import exceptions
from osc_lib import utils as oscutils

from manilaclient import api_versions
from manilaclient.common import cliutils
from manilaclient.osc.v2 import share_snapshots as osc_share_snapshots
from manilaclient.tests.unit.osc import osc_utils
from manilaclient.tests.unit.osc.v2 import fakes as manila_fakes

COLUMNS = ['ID', 'Name']

COLUMNS_DETAIL = [
    'ID',
    'Name',
    'Status',
    'Description',
    'Created At',
    'Size',
    'Share ID',
    'Share Proto',
    'Share Size',
    'User ID',
]


class TestShareSnapshot(manila_fakes.TestShare):
    def setUp(self):
        super().setUp()

        self.shares_mock = self.app.client_manager.share.shares
        self.shares_mock.reset_mock()

        self.snapshots_mock = self.app.client_manager.share.share_snapshots
        self.snapshots_mock.reset_mock()
        self.app.client_manager.share.api_version = api_versions.APIVersion(
            api_versions.MAX_VERSION
        )

        self.export_locations_mock = (
            self.app.client_manager.share.share_snapshot_export_locations
        )
        self.export_locations_mock.reset_mock()

        self.app.client_manager.share.api_version = api_versions.APIVersion(
            api_versions.MAX_VERSION
        )


class TestShareSnapshotCreate(TestShareSnapshot):
    def setUp(self):
        super().setUp()

        self.share = manila_fakes.FakeShare.create_one_share()
        self.shares_mock.create.return_value = self.share

        self.shares_mock.get.return_value = self.share

        self.share_snapshot = (
            manila_fakes.FakeShareSnapshot.create_one_snapshot(
                attrs={'status': 'available'}
            )
        )
        self.snapshots_mock.get.return_value = self.share_snapshot
        self.snapshots_mock.create.return_value = self.share_snapshot

        self.cmd = osc_share_snapshots.CreateShareSnapshot(self.app, None)

        self.data = tuple(self.share_snapshot._info.values())
        self.columns = tuple(self.share_snapshot._info.keys())

    def test_share_snapshot_create_missing_args(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            osc_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_share_snapshot_create_required_args(self):
        arglist = [self.share.id]
        verifylist = [('share', self.share.id)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.snapshots_mock.create.assert_called_with(
            share=self.share,
            force=False,
            name=None,
            description=None,
            metadata={},
        )

        self.assertCountEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_share_snapshot_create_force(self):
        arglist = [self.share.id, '--force']
        verifylist = [('share', self.share.id), ('force', True)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.snapshots_mock.create.assert_called_with(
            share=self.share,
            force=True,
            name=None,
            description=None,
            metadata={},
        )

        self.assertCountEqual(columns, columns)
        self.assertCountEqual(self.data, data)

    def test_share_snapshot_create(self):
        arglist = [
            self.share.id,
            '--name',
            self.share_snapshot.name,
            '--description',
            self.share_snapshot.description,
        ]
        verifylist = [
            ('share', self.share.id),
            ('name', self.share_snapshot.name),
            ('description', self.share_snapshot.description),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.snapshots_mock.create.assert_called_with(
            share=self.share,
            force=False,
            name=self.share_snapshot.name,
            description=self.share_snapshot.description,
            metadata={},
        )

        self.assertCountEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_share_snapshot_create_metadata(self):
        arglist = [
            self.share.id,
            '--name',
            self.share_snapshot.name,
            '--description',
            self.share_snapshot.description,
            '--property',
            'Manila=zorilla',
            '--property',
            'Zorilla=manila',
        ]
        verifylist = [
            ('share', self.share.id),
            ('name', self.share_snapshot.name),
            ('description', self.share_snapshot.description),
            ('property', {'Manila': 'zorilla', 'Zorilla': 'manila'}),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.snapshots_mock.create.assert_called_with(
            share=self.share,
            force=False,
            name=self.share_snapshot.name,
            description=self.share_snapshot.description,
            metadata={'Manila': 'zorilla', 'Zorilla': 'manila'},
        )

        self.assertCountEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_share_snapshot_create_wait(self):
        arglist = [self.share.id, '--wait']
        verifylist = [('share', self.share.id), ('wait', True)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.snapshots_mock.create.assert_called_with(
            share=self.share,
            force=False,
            name=None,
            description=None,
            metadata={},
        )

        self.snapshots_mock.get.assert_called_with(self.share_snapshot.id)
        self.assertCountEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    @mock.patch('manilaclient.osc.v2.share_snapshots.LOG')
    def test_share_snapshot_create_wait_error(self, mock_logger):
        arglist = [self.share.id, '--wait']
        verifylist = [('share', self.share.id), ('wait', True)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        with mock.patch('osc_lib.utils.wait_for_status', return_value=False):
            columns, data = self.cmd.take_action(parsed_args)

            self.snapshots_mock.create.assert_called_with(
                share=self.share,
                force=False,
                name=None,
                description=None,
                metadata={},
            )

            mock_logger.error.assert_called_with(
                "ERROR: Share snapshot is in error state."
            )

            self.snapshots_mock.get.assert_called_with(self.share_snapshot.id)
            self.assertCountEqual(self.columns, columns)
            self.assertCountEqual(self.data, data)


class TestShareSnapshotDelete(TestShareSnapshot):
    def setUp(self):
        super().setUp()

        self.share_snapshot = (
            manila_fakes.FakeShareSnapshot.create_one_snapshot()
        )

        self.snapshots_mock.get.return_value = self.share_snapshot

        self.cmd = osc_share_snapshots.DeleteShareSnapshot(self.app, None)

    def test_share_snapshot_delete_missing_args(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            osc_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_share_snapshot_delete(self):
        arglist = [self.share_snapshot.id]
        verifylist = [('snapshot', [self.share_snapshot.id])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.snapshots_mock.delete.assert_called_with(self.share_snapshot)
        self.assertIsNone(result)

    def test_share_snapshot_delete_force(self):
        arglist = [self.share_snapshot.id, '--force']
        verifylist = [('snapshot', [self.share_snapshot.id]), ('force', True)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.snapshots_mock.force_delete.assert_called_with(
            self.share_snapshot
        )
        self.assertIsNone(result)

    def test_share_snapshot_delete_multiple(self):
        share_snapshots = (
            manila_fakes.FakeShareSnapshot.create_share_snapshots(count=2)
        )
        arglist = [share_snapshots[0].id, share_snapshots[1].id]
        verifylist = [
            ('snapshot', [share_snapshots[0].id, share_snapshots[1].id])
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.assertEqual(
            self.snapshots_mock.delete.call_count, len(share_snapshots)
        )
        self.assertIsNone(result)

    def test_share_snapshot_delete_exception(self):
        arglist = [self.share_snapshot.id]
        verifylist = [('snapshot', [self.share_snapshot.id])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.snapshots_mock.delete.side_effect = exceptions.CommandError()
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )

    def test_share_snapshot_delete_wait(self):
        arglist = [self.share_snapshot.id, '--wait']
        verifylist = [('snapshot', [self.share_snapshot.id]), ('wait', True)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        with mock.patch('osc_lib.utils.wait_for_delete', return_value=True):
            result = self.cmd.take_action(parsed_args)

            self.snapshots_mock.delete.assert_called_with(self.share_snapshot)
            self.assertIsNone(result)

    def test_share_snapshot_delete_wait_error(self):
        arglist = [self.share_snapshot.id, '--wait']
        verifylist = [('snapshot', [self.share_snapshot.id]), ('wait', True)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        with mock.patch('osc_lib.utils.wait_for_delete', return_value=False):
            self.assertRaises(
                exceptions.CommandError, self.cmd.take_action, parsed_args
            )


class TestShareSnapshotShow(TestShareSnapshot):
    def setUp(self):
        super().setUp()

        self.export_location = (
            manila_fakes.FakeShareExportLocation.create_one_export_location()
        )

        self.share_snapshot = (
            manila_fakes.FakeShareSnapshot.create_one_snapshot(
                attrs={'export_locations': self.export_location}
            )
        )
        self.snapshots_mock.get.return_value = self.share_snapshot

        self.cmd = osc_share_snapshots.ShowShareSnapshot(self.app, None)

        self.data = self.share_snapshot._info.values()
        self.columns = self.share_snapshot._info.keys()
        self.export_locations_mock.list.return_value = [self.export_location]

    def test_share_snapshot_show_missing_args(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            osc_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_share_snapshot_show(self):
        arglist = [self.share_snapshot.id]
        verifylist = [('snapshot', self.share_snapshot.id)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        cliutils.convert_dict_list_to_string = mock.Mock()
        cliutils.convert_dict_list_to_string.return_value = (
            self.export_location
        )

        columns, data = self.cmd.take_action(parsed_args)
        self.snapshots_mock.get.assert_called_with(self.share_snapshot.id)
        self.assertCountEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)


class TestShareSnapshotSet(TestShareSnapshot):
    def setUp(self):
        super().setUp()

        self.share_snapshot = (
            manila_fakes.FakeShareSnapshot.create_one_snapshot(
                methods={"set_metadata": None}
            )
        )

        self.snapshots_mock.get.return_value = self.share_snapshot

        self.cmd = osc_share_snapshots.SetShareSnapshot(self.app, None)

    def test_set_snapshot_name(self):
        snapshot_name = 'snapshot-name-' + uuid.uuid4().hex
        arglist = [self.share_snapshot.id, '--name', snapshot_name]
        verifylist = [
            ('snapshot', self.share_snapshot.id),
            ('name', snapshot_name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.snapshots_mock.update.assert_called_with(
            self.share_snapshot, display_name=parsed_args.name
        )
        self.assertIsNone(result)

    def test_set_snapshot_description(self):
        description = 'snapshot-description-' + uuid.uuid4().hex
        arglist = [self.share_snapshot.id, '--description', description]
        verifylist = [
            ('snapshot', self.share_snapshot.id),
            ('description', description),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.snapshots_mock.update.assert_called_with(
            self.share_snapshot, display_description=parsed_args.description
        )
        self.assertIsNone(result)

    def test_set_snapshot_status(self):
        arglist = [self.share_snapshot.id, '--status', 'available']
        verifylist = [
            ('snapshot', self.share_snapshot.id),
            ('status', 'available'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.snapshots_mock.reset_state.assert_called_with(
            self.share_snapshot, parsed_args.status
        )
        self.assertIsNone(result)

    def test_set_snapshot_property(self):
        arglist = [
            self.share_snapshot.id,
            '--property',
            'Zorilla=manila',
        ]
        verifylist = [
            ('snapshot', self.share_snapshot.id),
            ('property', {'Zorilla': 'manila'}),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        self.share_snapshot.set_metadata.assert_called_with(
            {'Zorilla': 'manila'}
        )

    def test_set_snapshot_update_exception(self):
        snapshot_name = 'snapshot-name-' + uuid.uuid4().hex
        arglist = [self.share_snapshot.id, '--name', snapshot_name]
        verifylist = [
            ('snapshot', self.share_snapshot.id),
            ('name', snapshot_name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.snapshots_mock.update.side_effect = Exception()

        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )

    def test_set_snapshot_status_exception(self):
        arglist = [self.share_snapshot.id, '--status', 'available']
        verifylist = [
            ('snapshot', self.share_snapshot.id),
            ('status', 'available'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.snapshots_mock.reset_state.side_effect = Exception()

        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )

    def test_set_snapshot_property_exception(self):
        arglist = [
            '--property',
            'key=',
            self.share_snapshot.id,
        ]
        verifylist = [
            ('property', {'key': ''}),
            ('snapshot', self.share_snapshot.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        self.share_snapshot.set_metadata.assert_called_with({'key': ''})

        # '--property' takes key=value arguments
        # missing a value would raise a BadRequest
        self.share_snapshot.set_metadata.side_effect = exceptions.BadRequest
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )


class TestShareSnapshotUnset(TestShareSnapshot):
    def setUp(self):
        super().setUp()

        self.share_snapshot = (
            manila_fakes.FakeShareSnapshot.create_one_snapshot(
                methods={"delete_metadata": None}
            )
        )

        self.snapshots_mock.get.return_value = self.share_snapshot

        self.cmd = osc_share_snapshots.UnsetShareSnapshot(self.app, None)

    def test_unset_snapshot_name(self):
        arglist = [self.share_snapshot.id, '--name']
        verifylist = [('snapshot', self.share_snapshot.id), ('name', True)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.snapshots_mock.update.assert_called_with(
            self.share_snapshot, display_name=None
        )
        self.assertIsNone(result)

    def test_unset_snapshot_description(self):
        arglist = [self.share_snapshot.id, '--description']
        verifylist = [
            ('snapshot', self.share_snapshot.id),
            ('description', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.snapshots_mock.update.assert_called_with(
            self.share_snapshot, display_description=None
        )
        self.assertIsNone(result)

    def test_unset_snapshot_property(self):
        arglist = [
            '--property',
            'Manila',
            self.share_snapshot.id,
        ]
        verifylist = [
            ('property', ['Manila']),
            ('snapshot', self.share_snapshot.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        self.share_snapshot.delete_metadata.assert_called_with(
            parsed_args.property
        )

    def test_unset_snapshot_name_exception(self):
        arglist = [self.share_snapshot.id, '--name']
        verifylist = [('snapshot', self.share_snapshot.id), ('name', True)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.snapshots_mock.update.side_effect = Exception()

        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )

    def test_unset_snapshot_property_exception(self):
        arglist = [
            '--property',
            'Manila',
            self.share_snapshot.id,
        ]
        verifylist = [
            ('property', ['Manila']),
            ('snapshot', self.share_snapshot.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        self.share_snapshot.delete_metadata.assert_called_with(
            parsed_args.property
        )

        # 404 Not Found would be raised, if property 'Manila' doesn't exist
        self.share_snapshot.delete_metadata.side_effect = exceptions.NotFound
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )


@ddt.ddt
class TestShareSnapshotList(TestShareSnapshot):
    def setUp(self):
        super().setUp()

        self.share_snapshots = (
            manila_fakes.FakeShareSnapshot.create_share_snapshots(count=2)
        )
        self.snapshots_list = oscutils.sort_items(
            self.share_snapshots, 'name:asc', str
        )

        self.snapshots_mock.list.return_value = self.snapshots_list

        self.values = (
            oscutils.get_dict_properties(s._info, COLUMNS)
            for s in self.snapshots_list
        )

        self.cmd = osc_share_snapshots.ListShareSnapshot(self.app, None)

    def test_list_snapshots(self):
        arglist = []
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.snapshots_mock.list.assert_called_with(
            search_opts={
                'offset': None,
                'limit': None,
                'all_tenants': False,
                'name': None,
                'status': None,
                'share_id': None,
                'usage': None,
                'metadata': {},
                'name~': None,
                'description~': None,
                'description': None,
            }
        )

        self.assertEqual(COLUMNS, columns)
        self.assertEqual(list(self.values), list(data))

    def test_list_snapshots_all_projects(self):
        all_tenants_list = COLUMNS.copy()
        all_tenants_list.append('Project ID')
        list_values = (
            oscutils.get_dict_properties(s._info, all_tenants_list)
            for s in self.snapshots_list
        )

        arglist = ['--all-projects']

        verifylist = [('all_projects', True)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.snapshots_mock.list.assert_called_with(
            search_opts={
                'offset': None,
                'limit': None,
                'all_tenants': True,
                'name': None,
                'status': None,
                'share_id': None,
                'usage': None,
                'metadata': {},
                'name~': None,
                'description~': None,
                'description': None,
            }
        )

        self.assertEqual(all_tenants_list, columns)
        self.assertEqual(list(list_values), list(data))

    def test_list_snapshots_detail(self):
        values = (
            oscutils.get_dict_properties(s._info, COLUMNS_DETAIL)
            for s in self.snapshots_list
        )

        arglist = ['--detail']

        verifylist = [('detail', True)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.snapshots_mock.list.assert_called_with(
            search_opts={
                'offset': None,
                'limit': None,
                'all_tenants': False,
                'name': None,
                'status': None,
                'share_id': None,
                'usage': None,
                'metadata': {},
                'name~': None,
                'description~': None,
                'description': None,
            }
        )

        self.assertEqual(COLUMNS_DETAIL, columns)
        self.assertEqual(list(values), list(data))

    @ddt.data('2.35', '2.78')
    def test_list_snapshots_api_version_exception(self, v):
        self.app.client_manager.share.api_version = api_versions.APIVersion(v)

        if v == "2.35":
            arglist = ['--description', 'Description']
            verifylist = [('description', 'Description')]
        elif v == "2.78":
            arglist = [
                '--count',
            ]
            verifylist = [('count', True)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )

    def test_list_snapshots_share_id(self):
        self.share = manila_fakes.FakeShare.create_one_share(
            attrs={'id': self.snapshots_list[0].id}
        )

        self.shares_mock.get.return_value = self.share
        self.snapshots_mock.list.return_value = [self.snapshots_list[0]]

        values = (
            oscutils.get_dict_properties(s._info, COLUMNS)
            for s in [self.snapshots_list[0]]
        )

        arglist = ['--share', self.share.id]
        verifylist = [('share', self.share.id)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.snapshots_mock.list.assert_called_with(
            search_opts={
                'offset': None,
                'limit': None,
                'all_tenants': False,
                'name': None,
                'status': None,
                'share_id': self.share.id,
                'usage': None,
                'metadata': {},
                'name~': None,
                'description~': None,
                'description': None,
            }
        )

        self.assertEqual(COLUMNS, columns)
        self.assertEqual(list(values), list(data))

    def test_list_snapshots_with_count(self):
        self.app.client_manager.share.api_version = api_versions.APIVersion(
            '2.79'
        )

        self.snapshots_mock.list.return_value = self.snapshots_list, 2

        arglist = [
            '--count',
        ]
        verifylist = [('count', True)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.snapshots_mock.list.assert_called_with(
            search_opts={
                'offset': None,
                'limit': None,
                'all_tenants': False,
                'name': None,
                'status': None,
                'share_id': None,
                'usage': None,
                'metadata': {},
                'name~': None,
                'description~': None,
                'description': None,
                'with_count': True,
            }
        )


class TestShareSnapshotAdopt(TestShareSnapshot):
    def setUp(self):
        super().setUp()

        self.share = manila_fakes.FakeShare.create_one_share()
        self.shares_mock.get.return_value = self.share

        self.share_snapshot = (
            manila_fakes.FakeShareSnapshot.create_one_snapshot(
                attrs={'status': 'available'}
            )
        )

        self.snapshots_mock.get.return_value = self.share_snapshot
        self.export_location = (
            manila_fakes.FakeShareExportLocation.create_one_export_location()
        )

        self.snapshots_mock.manage.return_value = self.share_snapshot

        self.cmd = osc_share_snapshots.AdoptShareSnapshot(self.app, None)

        self.data = tuple(self.share_snapshot._info.values())
        self.columns = tuple(self.share_snapshot._info.keys())

    def test_share_snapshot_adopt_missing_args(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            osc_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_snapshot_adopt(self):
        arglist = [self.share.id, self.export_location.fake_path]
        verifylist = [
            ('share', self.share.id),
            ('provider_location', self.export_location.fake_path),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.snapshots_mock.manage.assert_called_with(
            share=self.share,
            provider_location=self.export_location.fake_path,
            driver_options={},
            name=None,
            description=None,
        )

        self.assertCountEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_snapshot_adopt_name(self):
        name = 'name-' + uuid.uuid4().hex
        arglist = [
            self.share.id,
            self.export_location.fake_path,
            '--name',
            name,
        ]
        verifylist = [
            ('share', self.share.id),
            ('provider_location', self.export_location.fake_path),
            ('name', name),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.snapshots_mock.manage.assert_called_with(
            share=self.share,
            provider_location=self.export_location.fake_path,
            driver_options={},
            name=name,
            description=None,
        )

        self.assertCountEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_snapshot_adopt_driver_option(self):
        arglist = [
            self.share.id,
            self.export_location.fake_path,
            '--driver-option',
            'key1=value1',
            '--driver-option',
            'key2=value2',
        ]
        verifylist = [
            ('share', self.share.id),
            ('provider_location', self.export_location.fake_path),
            ('driver_option', {'key1': 'value1', 'key2': 'value2'}),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.snapshots_mock.manage.assert_called_with(
            share=self.share,
            provider_location=self.export_location.fake_path,
            driver_options={'key1': 'value1', 'key2': 'value2'},
            name=None,
            description=None,
        )

        self.assertCountEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_snapshot_adopt_wait(self):
        arglist = [self.share.id, self.export_location.fake_path, '--wait']
        verifylist = [
            ('share', self.share.id),
            ('provider_location', self.export_location.fake_path),
            ('wait', True),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.snapshots_mock.get.assert_called_with(self.share_snapshot.id)
        self.snapshots_mock.manage.assert_called_with(
            share=self.share,
            provider_location=self.export_location.fake_path,
            driver_options={},
            name=None,
            description=None,
        )

        self.assertCountEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_snapshot_adopt_wait_error(self):
        arglist = [self.share.id, self.export_location.fake_path, '--wait']
        verifylist = [
            ('share', self.share.id),
            ('provider_location', self.export_location.fake_path),
            ('wait', True),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        with mock.patch('osc_lib.utils.wait_for_status', return_value=False):
            columns, data = self.cmd.take_action(parsed_args)
            self.snapshots_mock.get.assert_called_with(self.share_snapshot.id)
            self.snapshots_mock.manage.assert_called_with(
                share=self.share,
                provider_location=self.export_location.fake_path,
                driver_options={},
                name=None,
                description=None,
            )
            self.assertCountEqual(self.columns, columns)
            self.assertCountEqual(self.data, data)


class TestShareSnapshotAbandon(TestShareSnapshot):
    def setUp(self):
        super().setUp()

        self.share_snapshot = (
            manila_fakes.FakeShareSnapshot.create_one_snapshot(
                attrs={'status': 'available'}
            )
        )

        self.snapshots_mock.get.return_value = self.share_snapshot

        self.cmd = osc_share_snapshots.AbandonShareSnapshot(self.app, None)

    def test_share_snapshot_abandon_missing_args(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            osc_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_share_snapshot_abandon(self):
        arglist = [self.share_snapshot.id]
        verifylist = [('snapshot', [self.share_snapshot.id])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.snapshots_mock.unmanage.assert_called_with(self.share_snapshot)
        self.assertIsNone(result)

    def test_share_snapshot_abandon_multiple(self):
        share_snapshots = (
            manila_fakes.FakeShareSnapshot.create_share_snapshots(count=2)
        )
        arglist = [share_snapshots[0].id, share_snapshots[1].id]
        verifylist = [
            ('snapshot', [share_snapshots[0].id, share_snapshots[1].id])
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.assertEqual(
            self.snapshots_mock.unmanage.call_count, len(share_snapshots)
        )
        self.assertIsNone(result)

    def test_share_snapshot_abandon_wait(self):
        arglist = [self.share_snapshot.id, '--wait']
        verifylist = [('snapshot', [self.share_snapshot.id]), ('wait', True)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        with mock.patch('osc_lib.utils.wait_for_delete', return_value=True):
            result = self.cmd.take_action(parsed_args)
            self.snapshots_mock.unmanage.assert_called_with(
                self.share_snapshot
            )
            self.assertIsNone(result)

    def test_share_snapshot_abandon_wait_error(self):
        arglist = [self.share_snapshot.id, '--wait']
        verifylist = [('snapshot', [self.share_snapshot.id]), ('wait', True)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        with mock.patch('osc_lib.utils.wait_for_delete', return_value=False):
            self.assertRaises(
                exceptions.CommandError, self.cmd.take_action, parsed_args
            )


class TestShareSnapshotAccessAllow(TestShareSnapshot):
    def setUp(self):
        super().setUp()

        self.share_snapshot = (
            manila_fakes.FakeShareSnapshot.create_one_snapshot()
        )

        self.snapshots_mock.get.return_value = self.share_snapshot

        self.access_rule = (
            manila_fakes.FakeSnapshotAccessRule.create_one_access_rule()
        )
        self.snapshots_mock.allow.return_value = self.access_rule._info

        self.cmd = osc_share_snapshots.ShareSnapshotAccessAllow(self.app, None)

    def test_share_snapshot_access_allow(self):
        arglist = [self.share_snapshot.id, 'user', 'demo']
        verifylist = [
            ('snapshot', self.share_snapshot.id),
            ('access_type', 'user'),
            ('access_to', 'demo'),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.snapshots_mock.allow.assert_called_with(
            snapshot=self.share_snapshot, access_type='user', access_to='demo'
        )
        self.assertEqual(tuple(self.access_rule._info.keys()), columns)
        self.assertCountEqual(self.access_rule._info.values(), data)

    def test_share_snapshot_access_allow_exception(self):
        arglist = [self.share_snapshot.id, 'user', 'demo']
        verifylist = [
            ('snapshot', self.share_snapshot.id),
            ('access_type', 'user'),
            ('access_to', 'demo'),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.snapshots_mock.allow.side_effect = exceptions.CommandError()
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )


class TestShareSnapshotAccessDeny(TestShareSnapshot):
    def setUp(self):
        super().setUp()

        self.share_snapshot = (
            manila_fakes.FakeShareSnapshot.create_one_snapshot()
        )

        self.snapshots_mock.get.return_value = self.share_snapshot

        self.access_rule = (
            manila_fakes.FakeSnapshotAccessRule.create_one_access_rule()
        )

        self.cmd = osc_share_snapshots.ShareSnapshotAccessDeny(self.app, None)

    def test_share_snapshot_access_deny(self):
        arglist = [
            self.share_snapshot.id,
            self.access_rule.id,
        ]
        verifylist = [
            ('snapshot', self.share_snapshot.id),
            ('id', [self.access_rule.id]),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.snapshots_mock.deny.assert_called_with(
            snapshot=self.share_snapshot, id=self.access_rule.id
        )
        self.assertIsNone(result)

    def test_share_snapshot_access_deny_multiple(self):
        access_rules = manila_fakes.FakeSnapshotAccessRule.create_access_rules(
            count=2
        )

        arglist = [
            self.share_snapshot.id,
            access_rules[0].id,
            access_rules[1].id,
        ]
        verifylist = [
            ('snapshot', self.share_snapshot.id),
            ('id', [access_rules[0].id, access_rules[1].id]),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.assertEqual(
            self.snapshots_mock.deny.call_count, len(access_rules)
        )
        self.assertIsNone(result)

    def test_share_snapshot_access_deny_exception(self):
        arglist = [
            self.share_snapshot.id,
            self.access_rule.id,
        ]
        verifylist = [
            ('snapshot', self.share_snapshot.id),
            ('id', [self.access_rule.id]),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.snapshots_mock.deny.side_effect = exceptions.CommandError()
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )


class TestShareSnapshotAccessList(TestShareSnapshot):
    access_rules_columns = [
        'ID',
        'Access Type',
        'Access To',
        'State',
    ]

    def setUp(self):
        super().setUp()

        self.share_snapshot = (
            manila_fakes.FakeShareSnapshot.create_one_snapshot()
        )

        self.snapshots_mock.get.return_value = self.share_snapshot

        self.access_rules = (
            manila_fakes.FakeSnapshotAccessRule.create_access_rules(count=2)
        )

        self.snapshots_mock.access_list.return_value = self.access_rules
        self.cmd = osc_share_snapshots.ShareSnapshotAccessList(self.app, None)

        self.values = (
            oscutils.get_dict_properties(a._info, self.access_rules_columns)
            for a in self.access_rules
        )

    def test_share_snapshot_access_list(self):
        arglist = [self.share_snapshot.id]
        verifylist = [('snapshot', self.share_snapshot.id)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.snapshots_mock.access_list.assert_called_with(self.share_snapshot)

        self.assertEqual(self.access_rules_columns, columns)
        self.assertCountEqual(self.values, data)


class TestShareSnapshotExportLocationList(TestShareSnapshot):
    columns = ["ID", "Path"]

    def setUp(self):
        super().setUp()

        self.share_snapshot = (
            manila_fakes.FakeShareSnapshot.create_one_snapshot()
        )

        self.snapshots_mock.get.return_value = self.share_snapshot

        self.export_locations = (
            manila_fakes.FakeSnapshotExportLocation.create_export_locations()
        )

        self.export_locations_mock.list.return_value = self.export_locations
        self.values = (
            oscutils.get_dict_properties(e._info, self.columns)
            for e in self.export_locations
        )

        self.cmd = osc_share_snapshots.ShareSnapshotListExportLocation(
            self.app, None
        )

    def test_snapshot_export_locations_list(self):
        arglist = [self.share_snapshot.id]
        verifylist = [('snapshot', self.share_snapshot.id)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.export_locations_mock.list.assert_called_with(
            snapshot=self.share_snapshot
        )

        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.values, data)


class TestShareSnapshotExportLocationShow(TestShareSnapshot):
    def setUp(self):
        super().setUp()

        self.share_snapshot = (
            manila_fakes.FakeShareSnapshot.create_one_snapshot()
        )

        self.snapshots_mock.get.return_value = self.share_snapshot

        self.export_location = (
            manila_fakes.FakeSnapshotExportLocation.create_one_export_location()  # noqa E501
        )

        self.export_locations_mock.get.return_value = self.export_location

        self.cmd = osc_share_snapshots.ShareSnapshotShowExportLocation(
            self.app, None
        )

    def test_snapshot_export_locations_list(self):
        arglist = [self.share_snapshot.id, self.export_location.id]
        verifylist = [
            ('snapshot', self.share_snapshot.id),
            ('export_location', self.export_location.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.export_locations_mock.get.assert_called_with(
            export_location=self.export_location.id,
            snapshot=self.share_snapshot,
        )

        self.assertEqual(tuple(self.export_location._info.keys()), columns)
        self.assertCountEqual(self.export_location._info.values(), data)



# Copyright (c) 2025 Cloudification GmbH.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import json

from manilaclient.tests.functional.osc import base


class QosTypesCLITest(base.OSCClientTestBase):
    def test_qos_type_create(self):
        name = 'test_qos_type'
        description = 'Description'
        qos_type = self.create_qos_type(name=name, description=description)

        self.assertEqual(name, qos_type["name"])
        self.assertEqual(description, qos_type["description"])

        qos_types_list = self.listing_result('share qos type', 'list')
        self.assertIn(qos_type["id"], [item['ID'] for item in qos_types_list])

    def test_qos_type_create_specs(self):
        name = 'test_qos_type2'
        qos_type = self.create_qos_type(
            name=name,
            specs={"foo": "bar", "manila": "gazpacho"},
            formatter='json',
        )

        specs = qos_type["specs"]
        self.assertEqual("bar", specs["foo"])
        self.assertEqual("gazpacho", specs["manila"])

    def test_qos_type_create_specs_using_command(self):
        self.openstack(
            'share qos type create test_qos_type2_cli '
            '--description test_cli --spec foo1=bar1 --spec foo2=bar2'
        )
        self.addCleanup(
            self.openstack, 'share qos type delete test_qos_type2_cli'
        )

        qos_type = json.loads(
            self.openstack('share qos type show test_qos_type2_cli -f json')
        )

        self.assertEqual('test_cli', qos_type["description"])
        self.assertEqual('bar1', qos_type["specs"]["foo1"])
        self.assertEqual('bar2', qos_type["specs"]["foo2"])

    def test_qos_type_delete(self):
        qos_type_1 = self.create_qos_type(
            name='test_qos_type3', add_cleanup=False
        )
        qos_type_2 = self.create_qos_type(
            name='test_qos_type4', add_cleanup=False
        )

        self.openstack(
            f'share qos type delete {qos_type_1["id"]} {qos_type_2["id"]}'
        )

        self.check_object_deleted('share qos type', qos_type_1["id"])
        self.check_object_deleted('share qos type', qos_type_2["id"])

    def test_qos_type_set(self):
        qos_type = self.create_qos_type(name='test_qos_type5')

        self.openstack(
            f'share qos type set {qos_type["id"]} --description Description'
            ' --spec foo=bar2'
        )

        qos_type = json.loads(
            self.openstack(f'share qos type show {qos_type["id"]} -f json')
        )

        self.assertEqual('Description', qos_type["description"])
        self.assertEqual('bar2', qos_type["specs"]["foo"])

    def test_qos_type_unset(self):
        qos_type = self.create_qos_type(
            name='test_qos_type6', specs={'foo': 'bar', 'foo1': 'bar1'}
        )

        self.openstack(
            f'share qos type unset {qos_type["id"]} --spec foo --spec foo1'
        )

        qos_type = json.loads(
            self.openstack(f'share qos type show {qos_type["id"]} -f json')
        )

        self.assertNotIn('foo', qos_type["specs"])
        self.assertNotIn('foo1', qos_type["specs"])

    def test_qos_type_list(self):
        qos_type_1 = self.create_qos_type(name='test_qos_type7')
        qos_type_2 = self.create_qos_type(
            name='test_qos_type8', specs={'foo': 'bar'}
        )

        types_list = self.listing_result(
            'share qos type', 'list', client=self.admin_client
        )

        self.assertTableStruct(
            types_list,
            [
                'ID',
                'Name',
                'Description',
                'Specs',
            ],
        )
        id_list = [item['ID'] for item in types_list]
        self.assertIn(qos_type_1['id'], id_list)
        self.assertIn(qos_type_2['id'], id_list)

        types_list = self.listing_result('share qos type', 'list')

        id_list = [item['ID'] for item in types_list]
        self.assertIn(qos_type_1['id'], id_list)
        self.assertIn(qos_type_2['id'], id_list)

    def test_qos_type_show(self):
        qos_type = self.create_qos_type(
            name='test_qos_type10', specs={'foo': 'bar'}
        )

        result = json.loads(
            self.openstack(f'share qos type show {qos_type["id"]} -f json')
        )

        self.assertEqual(qos_type["name"], result["name"])
        self.assertEqual('bar', result["specs"]["foo"])



# Copyright 2010 Jacob Kaplan-Moss

# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Base utilities to build API operation managers and objects on top of.
"""

import abc
import contextlib
import copy
import hashlib
import os

from oslo_utils import strutils

from manilaclient import api_versions
from manilaclient.common import cliutils
from manilaclient import exceptions
from manilaclient import utils


def getid(obj):
    """Return id if argument is a Resource.

    Abstracts the common pattern of allowing both an object or an object's ID
    (UUID) as a parameter when dealing with relationships.
    """
    try:
        if obj.uuid:
            return obj.uuid
    except AttributeError:
        pass
    try:
        return obj.id
    except AttributeError:
        return obj


class Manager(utils.HookableMixin):
    """Manager for CRUD operations.

    Managers interact with a particular type of API (shares, snapshots,
    etc.) and provide CRUD operations for them.
    """

    resource_class = None

    def __init__(self, api):
        self.api = api
        self.client = api.client

    @property
    def api_version(self):
        return self.api.api_version

    def _list(
        self, url, response_key, manager=None, body=None, return_raw=None
    ):
        """List the collection.

        :param url: a partial URL, e.g., '/shares'
        :param response_key: the key to be looked up in response dictionary,
            e.g., 'shares'. If response_key is None - all response body
            will be used.
        :param manager: manager instance for constructing the returned objects
            (self will be used by default)
        :param body: data that will be encoded as JSON and passed in POST
            request (GET will be sent by default)
        """
        resp = None
        if body:
            resp, body = self.api.client.post(url, body=body)
        else:
            resp, body = self.api.client.get(url)

        if manager is None:
            manager = self

        obj_class = manager.resource_class

        data = body[response_key]
        # NOTE(ja): keystone returns values as list as {'values': [ ... ]}
        #           unlike other services which just return the list...
        if isinstance(data, dict):
            try:
                data = data['values']
            except KeyError:
                pass
        with self.completion_cache('human_id', obj_class, mode="w"):
            with self.completion_cache('uuid', obj_class, mode="w"):
                if return_raw:
                    return data
                resource = [
                    obj_class(manager, res, loaded=True) for res in data if res
                ]
                if 'count' in body:
                    return resource, body['count']
                else:
                    return resource

    @contextlib.contextmanager
    def completion_cache(self, cache_type, obj_class, mode):
        """Bash autocompletion items storage.

        The completion cache store items that can be used for bash
        autocompletion, like UUIDs or human-friendly IDs.

        A resource listing will clear and repopulate the cache.

        A resource create will append to the cache.

        Delete is not handled because listings are assumed to be performed
        often enough to keep the cache reasonably up-to-date.
        """
        base_dir = cliutils.env(
            'manilaclient_UUID_CACHE_DIR',
            'MANILACLIENT_UUID_CACHE_DIR',
            default="~/.cache/manilaclient",
        )

        # NOTE(sirp): Keep separate UUID caches for each username + endpoint
        # pair
        username = cliutils.env('OS_USERNAME', 'MANILA_USERNAME')
        url = cliutils.env('OS_URL', 'MANILA_URL')
        uniqifier = hashlib.sha256(
            username.encode('utf-8') + url.encode('utf-8')
        ).hexdigest()

        cache_dir = os.path.expanduser(os.path.join(base_dir, uniqifier))

        try:
            os.makedirs(cache_dir, 0o755)
        except OSError:
            # NOTE(kiall): This is typically either permission denied while
            #              attempting to create the directory, or the directory
            #              already exists. Either way, don't fail.
            pass

        resource = obj_class.__name__.lower()
        filename = "{}-{}-cache".format(resource, cache_type.replace('_', '-'))
        path = os.path.join(cache_dir, filename)

        cache_attr = f"_{cache_type}_cache"

        try:
            setattr(self, cache_attr, open(path, mode))
        except OSError:
            # NOTE(kiall): This is typically a permission denied while
            #              attempting to write the cache file.
            pass

        try:
            yield
        finally:
            cache = getattr(self, cache_attr, None)
            if cache:
                cache.close()
                delattr(self, cache_attr)

    def write_to_completion_cache(self, cache_type, val):
        cache = getattr(self, f"_{cache_type}_cache", None)
        if cache:
            try:
                cache.write(f"{val}\n")
            except UnicodeEncodeError:
                pass

    def _get(self, url, response_key, return_raw=False):
        resp, body = self.api.client.get(url)
        if response_key:
            if return_raw:
                return body[response_key]
            return self.resource_class(self, body[response_key], loaded=True)
        else:
            return self.resource_class(self, body, loaded=True)

    def _get_with_base_url(self, url, response_key=None):
        resp, body = self.api.client.get_with_base_url(url)
        if response_key:
            return [
                self.resource_class(self, res, loaded=True)
                for res in body[response_key]
                if res
            ]
        else:
            return self.resource_class(self, body, loaded=True)

    def _create(self, url, body, response_key, return_raw=False, **kwargs):
        self.run_hooks('modify_body_for_create', body, **kwargs)
        resp, body = self.api.client.post(url, body=body)
        if return_raw:
            return body[response_key]

        with self.completion_cache('human_id', self.resource_class, mode="a"):
            with self.completion_cache('uuid', self.resource_class, mode="a"):
                return self.resource_class(self, body[response_key])

    def _accept(self, url, body):
        resp, body = self.api.client.post(url, body=body)

    def _delete(self, url):
        resp, body = self.api.client.delete(url)

    def _update(self, url, body, response_key=None, **kwargs):
        self.run_hooks('modify_body_for_update', body, **kwargs)
        resp, body = self.api.client.put(url, body=body)
        if body:
            if response_key:
                return self.resource_class(self, body[response_key])
            else:
                return self.resource_class(self, body)

    def _build_query_string(self, search_opts):
        search_opts = search_opts or {}
        params = sorted([(k, v) for (k, v) in search_opts.items() if v])
        query_string = f"?{utils.safe_urlencode(params)}" if params else ''
        return query_string


class ManagerWithFind(Manager):
    """Like a `Manager`, but with additional `find()`/`findall()` methods."""

    def find(self, **kwargs):
        """Find a single item with attributes matching ``**kwargs``.

        This isn't very efficient: it loads the entire list then filters on
        the Python side.
        """
        matches = self.findall(**kwargs)
        num_matches = len(matches)
        if num_matches == 0:
            msg = f"No {self.resource_class.__name__} matching {kwargs}."
            raise exceptions.NotFound(404, msg)
        elif num_matches > 1:
            raise exceptions.NoUniqueMatch
        else:
            return matches[0]

    def findall(self, **kwargs):
        """Find all items with attributes matching ``**kwargs``.

        This isn't very efficient: it loads the entire list then filters on
        the Python side.
        """
        found = []
        searches = list(kwargs.items())

        search_opts = {'all_tenants': 1}
        resources = self.list(search_opts=search_opts)
        if 'v2.shares.ShareManager' in str(
            self.__class__
        ) and self.api_version >= api_versions.APIVersion("2.69"):
            search_opts_2 = {'all_tenants': 1, 'is_soft_deleted': True}
            shares_soft_deleted = self.list(search_opts=search_opts_2)
            resources += shares_soft_deleted
        for obj in resources:
            try:
                if all(
                    getattr(obj, attr) == value for (attr, value) in searches
                ):
                    found.append(obj)
            except AttributeError:
                continue

        return found

    def list(self, search_opts=None):
        raise NotImplementedError


class Resource:
    """Base class for OpenStack resources (tenant, user, etc.).

    This is pretty much just a bag for attributes.
    """

    HUMAN_ID = False
    NAME_ATTR = 'name'

    def __init__(self, manager, info, loaded=False):
        """Populate and bind to a manager.

        :param manager: BaseManager object
        :param info: dictionary representing resource attributes
        :param loaded: prevent lazy-loading if set to True
        """
        self.manager = manager
        self._info = info
        self._add_details(info)
        self._loaded = loaded

    def __repr__(self):
        reprkeys = sorted(
            k for k in self.__dict__.keys() if k[0] != '_' and k != 'manager'
        )
        info = ", ".join(f"{k}={getattr(self, k)}" for k in reprkeys)
        return f"<{self.__class__.__name__} {info}>"

    @property
    def human_id(self):
        """Human-readable ID which can be used for bash completion."""
        if self.HUMAN_ID:
            name = getattr(self, self.NAME_ATTR, None)
            if name is not None:
                return strutils.to_slug(name)
        return None

    def _add_details(self, info):
        for k, v in info.items():
            try:
                setattr(self, k, v)
                self._info[k] = v
            except AttributeError:
                # In this case we already defined the attribute on the class
                pass

    def __getattr__(self, k):
        if k not in self.__dict__:
            # NOTE(bcwaldon): disallow lazy-loading if already loaded once
            if not self.is_loaded():
                self.get()
                return self.__getattr__(k)

            raise AttributeError(k)
        else:
            return self.__dict__[k]

    def get(self):
        """Support for lazy loading details.

        Some clients, such as novaclient have the option to lazy load the
        details, details which can be loaded with this function.
        """
        # set_loaded() first ... so if we have to bail, we know we tried.
        self.set_loaded(True)
        if not hasattr(self.manager, 'get'):
            return

        new = self.manager.get(self.id)
        if new:
            self._add_details(new._info)

    def __eq__(self, other):
        if not isinstance(other, Resource):
            return NotImplemented
        # two resources of different types are not equal
        if not isinstance(other, self.__class__):
            return False
        if hasattr(self, 'id') and hasattr(other, 'id'):
            return self.id == other.id
        return self._info == other._info

    def __ne__(self, other):
        return not self == other

    def is_loaded(self):
        return self._loaded

    def set_loaded(self, val):
        self._loaded = val

    def to_dict(self):
        return copy.deepcopy(self._info)


class MetadataCapableResource(Resource, metaclass=abc.ABCMeta):
    superresource = None

    def _get_subresource_and_resource(self, superresource):
        resource = self
        subresource = None
        superresource = superresource or self.superresource
        if superresource is not None:
            resource = superresource
            subresource = self
        return resource, subresource

    def get_metadata(self, superresource=None):
        """Get metadata of a resource

        :param superresource: either a parent resource object or text with
            its ID. Required for sub-resources such as share export
            locations which do not include a reference to the parent object
            by default
        """

        resource, subresource = self._get_subresource_and_resource(
            superresource
        )

        return self.manager.get_metadata(resource, subresource=subresource)

    def set_metadata(self, metadata, superresource=None):
        """Set or update metadata for the resource.

        :param metadata: A dictionary of key:value pairs to be set as
            resource metadata
        :param superresource: either a parent resource object or text with
            its ID. Required for sub-resources such as share share export
            locations which do not include a reference to the parent object
            by default
        """
        resource, subresource = self._get_subresource_and_resource(
            superresource
        )

        return self.manager.set_metadata(
            resource, metadata, subresource=subresource
        )

    def delete_metadata(self, keys, superresource=None):
        """Delete specified keys from the given resource.

        :param keys: An iterable with keys of metadata items to be deleted
        :param superresource: either a parent resource object or text with
            its ID. Required for sub-resources such as share share export
            locations which do not include a reference to the parent object
            by default
        """
        resource, subresource = self._get_subresource_and_resource(
            superresource
        )

        return self.manager.delete_metadata(
            resource, keys, subresource=subresource
        )

    def update_all_metadata(self, metadata, superresource=None):
        """Update all metadata for this resource.

        :param metadata: A dictionary of key:value pairs of resource metadata
            to be updated
        :param superresource: either a parent resource object or text with
            its ID. Required for sub-resources such as share share export
            locations which do not include a reference to the parent object
            by default
        """
        resource, subresource = self._get_subresource_and_resource(
            superresource
        )

        return self.manager.update_all_metadata(
            resource, metadata, subresource=subresource
        )


class MetadataCapableManager(ManagerWithFind, metaclass=abc.ABCMeta):
    """Provides extended behavior to objects to handle key=value metadata."""

    resource_path = None
    subresource_path = None

    def get_metadata(self, resource, subresource=None):
        """Get metadata of a resource.

        :param resource: either resource object or text with its ID.
        :param subresource: either a child resource object or text with its ID
        """
        resource = getid(resource)
        if subresource:
            subresource = getid(subresource)
            resource = f"{resource}{self.subresource_path}/{subresource}"

        return self._get(
            f"{self.resource_path}/{resource}/metadata", "metadata"
        )

    def set_metadata(self, resource, metadata, subresource=None):
        """Set or update metadata for resource.

        :param resource: either resource object or text with its ID.
        :param metadata: A dictionary of key:value pairs to be set as
            resource metadata
        :param subresource: either a child resource object or text with its ID
        """
        body = {'metadata': metadata}
        resource = getid(resource)
        if subresource:
            subresource = getid(subresource)
            resource = f"{resource}{self.subresource_path}/{subresource}"

        return self._create(
            f"{self.resource_path}/{resource}/metadata", body, "metadata"
        )

    def delete_metadata(self, resource, keys, subresource=None):
        """Delete specified keys from resource metadata.

        :param resource: either resource object or text with its ID.
        :param keys: An iterable with keys of metadata items to be deleted
        :param subresource: either a child resource object or text with its ID
        """
        resource = getid(resource)
        if subresource:
            subresource = getid(subresource)
            resource = f"{resource}{self.subresource_path}/{subresource}"

        for key in keys:
            self._delete(f"{self.resource_path}/{resource}/metadata/{key}")

    def update_all_metadata(self, resource, metadata, subresource=None):
        """Update all metadata of a resource.

        :param resource: either resource object or text with its ID.
        :param metadata: A dictionary of key:value pairs of resource metadata
            to be updated
        :param subresource: either a child resource object or text with its ID
        """
        body = {'metadata': metadata}
        resource = getid(resource)
        if subresource:
            subresource = getid(subresource)
            resource = f"{resource}{self.subresource_path}/{subresource}"

        return self._update(f"{self.resource_path}/{resource}/metadata", body)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import logging

from openstackclient.identity import common as identity_common
from osc_lib.cli import parseractions
from osc_lib.command import command
from osc_lib import exceptions
from osc_lib import utils as osc_utils

from manilaclient import api_versions
from manilaclient.common._i18n import _
from manilaclient.common.apiclient import utils as apiutils
from manilaclient.common import cliutils
from manilaclient.common import constants

LOG = logging.getLogger(__name__)


class DeleteShareServer(command.Command):
    """Delete one or more share servers (Admin only)"""

    _description = _("Delete one or more share servers")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "share_servers",
            metavar="<share-server>",
            nargs="+",
            help=_("ID(s) of the server(s) to delete"),
        )
        parser.add_argument(
            "--wait",
            action='store_true',
            default=False,
            help=_("Wait for share server deletion."),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share
        result = 0

        for server in parsed_args.share_servers:
            try:
                server_obj = osc_utils.find_resource(
                    share_client.share_servers, server
                )

                share_client.share_servers.delete(server_obj)
                if parsed_args.wait:
                    if not osc_utils.wait_for_delete(
                        manager=share_client.share_servers,
                        res_id=server_obj.id,
                    ):
                        result += 1

            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to delete a share server with "
                        "ID '%(server)s': %(e)s"
                    ),
                    {'server': server, 'e': e},
                )

        if result > 0:
            msg = "Failed to delete %(result)d servers out %(total)d"
            raise exceptions.CommandError(
                msg
                % {'result': result, 'total': len(parsed_args.share_servers)},
            )


class ShowShareServer(command.ShowOne):
    """Show share server (Admin only)."""

    _description = _("Show details about a share server (Admin only).")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "share_server",
            metavar="<share-server>",
            help=_("ID of share server."),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        share_server = osc_utils.find_resource(
            share_client.share_servers, parsed_args.share_server
        )

        # All 'backend_details' data already present as separated strings,
        # so remove big dict from view.
        if "backend_details" in share_server._info:
            del share_server._info["backend_details"]

        share_server._info.pop('links', None)
        return self.dict2columns(share_server._info)


class ListShareServer(command.Lister):
    """List all share servers (Admin only)."""

    _description = _("List all share servers (Admin only).")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--host',
            metavar='<hostname>',
            default=None,
            help=_('Filter results by name of host.'),
        )
        parser.add_argument(
            '--status',
            metavar="<status>",
            default=None,
            help=_('Filter results by status.'),
        )
        parser.add_argument(
            '--share-network',
            metavar='<share-network>',
            default=None,
            help=_('Filter results by share network name or ID.'),
        )
        parser.add_argument(
            '--project',
            metavar='<project>',
            default=None,
            help=_('Filter results by project name or ID.'),
        )
        parser.add_argument(
            '--share-network-subnet',
            metavar='<share-network-subnet>',
            type=str,
            default=None,
            help=_(
                "Filter results by share network subnet that the "
                "share server's network allocation exists within. "
                "Available for microversion >= 2.51 (Optional, "
                "Default=None)"
            ),
        )
        parser.add_argument(
            '--source-share-server-id',
            metavar='<source-share-server-id>',
            type=str,
            default=None,
            help=_(
                "Share server ID to be used as a filter. Available for "
                "microversion >= 2.57 (Optional, Default=None)"
            ),
        )
        parser.add_argument(
            '--identifier',
            metavar='<identifier>',
            type=str,
            default=None,
            help=_(
                "Identifier of the share server in the share back end. "
                "Available for microversion >= 2.49 "
                "(Optional, Default=None)"
            ),
        )
        identity_common.add_project_domain_option_to_parser(parser)
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share
        identity_client = self.app.client_manager.identity

        project_id = None
        if parsed_args.project:
            project_id = identity_common.find_project(
                identity_client,
                parsed_args.project,
                parsed_args.project_domain,
            ).id

        if (
            parsed_args.identifier
            and share_client.api_version < api_versions.APIVersion("2.49")
        ):
            raise exceptions.CommandError(
                "Filtering by identifier is only allowed with manila API "
                "version >= 2.49."
            )
        if (
            parsed_args.share_network_subnet
            and share_client.api_version < api_versions.APIVersion("2.51")
        ):
            raise exceptions.CommandError(
                "Share network subnet can be specified only with manila API "
                "version >= 2.51"
            )
        if (
            parsed_args.source_share_server_id
            and share_client.api_version < api_versions.APIVersion("2.57")
        ):
            raise exceptions.CommandError(
                "Filtering by source_share_server_id is only allowed with "
                "manila API version >= 2.57."
            )

        columns = [
            'ID',
            'Host',
            'Status',
            'Share Network ID',
            'Project ID',
        ]

        search_opts = {
            'status': parsed_args.status,
            'host': parsed_args.host,
            'project_id': project_id,
        }

        if parsed_args.share_network:
            share_network_id = osc_utils.find_resource(
                share_client.share_networks, parsed_args.share_network
            ).id
            search_opts['share_network'] = share_network_id

        if parsed_args.source_share_server_id:
            search_opts['source_share_server_id'] = (
                parsed_args.source_share_server_id
            )

        if parsed_args.identifier:
            search_opts['identifier'] = parsed_args.identifier

        if parsed_args.share_network_subnet:
            search_opts['share_network_subnet_id'] = (
                parsed_args.share_network_subnet
            )

        share_servers = share_client.share_servers.list(
            search_opts=search_opts
        )

        data = (
            osc_utils.get_dict_properties(share_server._info, columns)
            for share_server in share_servers
        )

        return (columns, data)


class AdoptShareServer(command.ShowOne):
    """Adopt share server not handled by Manila (Admin only)."""

    _description = _("Adopt share server not handled by Manila (Admin only).")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'host',
            metavar='<host>',
            type=str,
            help=_('Backend name as "<node_hostname>@<backend_name>".'),
        )
        parser.add_argument(
            "share_network",
            metavar="<share-network>",
            help=_(
                "Share network where share server has network allocations in."
            ),
        )
        parser.add_argument(
            'identifier',
            metavar='<identifier>',
            type=str,
            help=_(
                "A driver-specific share server identifier required "
                "by the driver to manage the share server."
            ),
        )
        parser.add_argument(
            '--driver-options',
            metavar='<key=value>',
            action=parseractions.KeyValueAction,
            default={},
            help=_(
                "One or more driver-specific key=value pairs that may be "
                "necessary to manage the share server (Optional, "
                "Default=None)."
            ),
        )
        parser.add_argument(
            '--share-network-subnet',
            type=str,
            metavar='<share-network-subnet>',
            default=None,
            help="Share network subnet where share server has network  "
            "allocations in.The default subnet will be used if "
            "it's not specified. Available for microversion "
            ">= 2.51 (Optional, Default=None).",
        )
        parser.add_argument(
            "--wait",
            action='store_true',
            help=_("Wait until share server is adopted"),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        share_network = None
        if parsed_args.share_network:
            share_network = osc_utils.find_resource(
                share_client.share_networks, parsed_args.share_network
            ).id

        share_network_subnet = None
        if (
            parsed_args.share_network_subnet
            and share_client.api_version < api_versions.APIVersion("2.51")
        ):
            raise exceptions.CommandError(
                "Share network subnet can be specified only with manila API "
                "version >= 2.51"
            )
        elif parsed_args.share_network_subnet:
            share_network_subnet = share_client.share_network_subnets.get(
                share_network, parsed_args.share_network_subnet
            ).id

        share_server = share_client.share_servers.manage(
            host=parsed_args.host,
            share_network_id=share_network,
            identifier=parsed_args.identifier,
            driver_options=parsed_args.driver_options,
            share_network_subnet_id=share_network_subnet,
        )

        if parsed_args.wait:
            if not osc_utils.wait_for_status(
                status_f=share_client.share_servers.get,
                res_id=share_server.id,
                success_status=['active'],
                error_status=['manage_error', 'error'],
            ):
                LOG.error(_("ERROR: Share server is in error state."))

            share_server = osc_utils.find_resource(
                share_client.share_servers, share_server.id
            )

        share_server._info.pop('links', None)

        # All 'backend_details' data already present as separated strings,
        # so remove big dict from view.
        share_server._info.pop("backend_details", None)

        return self.dict2columns(share_server._info)


class AbandonShareServer(command.Command):
    """Remove one or more share servers (Admin only)."""

    _description = _("Remove one or more share server(s) (Admin only).")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "share_server",
            metavar="<share-server>",
            nargs='+',
            help=_("ID of the server(s) to be abandoned."),
        )
        parser.add_argument(
            "--force",
            action='store_true',
            default=False,
            help=_(
                "Enforces the unmanage share server operation, even "
                "if the backend driver does not support it."
            ),
        )
        parser.add_argument(
            "--wait",
            action='store_true',
            default=False,
            help=_("Wait until share server is abandoned"),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share
        result = 0

        for server in parsed_args.share_server:
            try:
                server_obj = osc_utils.find_resource(
                    share_client.share_servers, server
                )
                kwargs = {}
                if parsed_args.force:
                    kwargs['force'] = parsed_args.force
                share_client.share_servers.unmanage(server_obj, **kwargs)

                if parsed_args.wait:
                    if not osc_utils.wait_for_delete(
                        manager=share_client.share_servers,
                        res_id=server_obj.id,
                    ):
                        result += 1

            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to abandon share server with "
                        "ID '%(server)s': %(e)s"
                    ),
                    {'server': server, 'e': e},
                )

        if result > 0:
            msg = _('Failed to abandon %(result)s of %(total)s servers.')
            raise exceptions.CommandError(
                msg
                % {'result': result, 'total': len(parsed_args.share_server)}
            )


class SetShareServer(command.Command):
    """Set share server properties."""

    _description = _("Set share server properties (Admin only).")

    def get_parser(self, prog_name):
        allowed_update_choices = [
            'unmanage_starting',
            'server_migrating_to',
            'error',
            'unmanage_error',
            'manage_error',
            'inactive',
            'active',
            'server_migrating',
            'manage_starting',
            'deleting',
            'network_change',
        ]
        allowed_update_choices_str = ', '.join(allowed_update_choices)

        parser = super().get_parser(prog_name)
        parser.add_argument(
            "share_server",
            metavar="<share-server>",
            help=_("ID of the share server to modify."),
        )
        parser.add_argument(
            "--status",
            metavar="<status>",
            required=False,
            default=constants.STATUS_ACTIVE,
            help=_(
                "Assign a status to the share server. Options "
                "include: %s. If no state is "
                "provided, active will be "
                "used."
            )
            % allowed_update_choices_str,
        )
        parser.add_argument(
            '--task-state',
            metavar="<task-state>",
            required=False,
            default=None,
            nargs='?',
            help=_(
                "Indicate which task state to assign the share server. "
                "Options include migration_starting, migration_in_progress,"
                " migration_completing, migration_success, migration_error,"
                " migration_cancelled, migration_driver_in_progress, "
                "migration_driver_phase1_done, data_copying_starting, "
                "data_copying_in_progress, data_copying_completing, "
                "data_copying_completed, data_copying_cancelled, "
                "data_copying_error. "
            ),
        )
        return parser

    def take_action(self, parsed_args):
        if not parsed_args.status and not parsed_args.task_state:
            msg = _(
                "A status or a task state should be provided for this command."
            )
            LOG.error(msg)
            raise exceptions.CommandError(msg)
        share_client = self.app.client_manager.share

        share_server = osc_utils.find_resource(
            share_client.share_servers, parsed_args.share_server
        )

        if parsed_args.status:
            try:
                share_client.share_servers.reset_state(
                    share_server, parsed_args.status
                )
            except Exception as e:
                msg = (
                    _("Failed to set status '%(status)s': %(exception)s"),
                    {'status': parsed_args.status, 'exception': e},
                )
                LOG.error(msg)
                raise exceptions.CommandError(msg)

        if hasattr(parsed_args, 'task_state'):
            if share_client.api_version < api_versions.APIVersion("2.57"):
                raise exceptions.CommandError(
                    "Setting the state of a share server is only available "
                    "with manila API version >= 2.57"
                )
            else:
                task_state = parsed_args.task_state
                if task_state and task_state.lower() == "none":
                    task_state = None
                result = 0
                try:
                    share_client.share_servers.reset_task_state(
                        share_server, task_state
                    )
                except Exception as e:
                    LOG.error(
                        _("Failed to update share server task state %s"), e
                    )
                    result += 1

            if result > 0:
                raise exceptions.CommandError(
                    _("One or more of the reset operations failed")
                )


class ShareServerMigrationCancel(command.Command):
    """Attempts to cancel migration for a given share server

    :param share_server: either share_server object or text with its ID.

    """

    _description = _("Cancels migration of a given share server when copying")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'share_server',
            metavar='<share_server>',
            help=_('ID of share server to cancel migration.'),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share
        share_server = osc_utils.find_resource(
            share_client.share_servers, parsed_args.share_server
        )
        if share_client.api_version >= api_versions.APIVersion("2.57"):
            share_server.migration_cancel()
        else:
            raise exceptions.CommandError(
                "Share Server Migration cancel is only available "
                "with manila API version >= 2.57"
            )


class ShareServerMigrationComplete(command.Command):
    """Completes migration for a share server (Admin only, Experimental)."""

    _description = _("Completes migration for a given share server")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'share_server',
            metavar='<share_server>',
            help=_('ID of share server to complete migration.'),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share
        share_server = osc_utils.find_resource(
            share_client.share_servers, parsed_args.share_server
        )
        if share_client.api_version >= api_versions.APIVersion("2.57"):
            share_server.migration_complete()
        else:
            raise exceptions.CommandError(
                "Share Server Migration complete is only available "
                "with manila API version >= 2.57"
            )


class ShareServerMigrationShow(command.ShowOne):
    """Obtains progress of share migration for a given share server.

    (Admin only, Experimental).

    :param share_server: either share_server object or text with its ID.

    """

    _description = _(
        "Gets migration progress of a given share server when copying"
    )

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'share_server',
            metavar='<share_server>',
            help='ID of share server to show migration progress for.',
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share
        if share_client.api_version >= api_versions.APIVersion("2.57"):
            result = share_client.share_servers.migration_get_progress(
                parsed_args.share_server
            )
            return self.dict2columns(result)
        else:
            raise exceptions.CommandError(
                "Share Server Migration show is only available "
                "with manila API version >= 2.57"
            )


class ShareServerMigrationStart(command.ShowOne):
    """Migrates share server to a new host (Admin only, Experimental)."""

    _description = _("Migrates share server to a new host.")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'share_server',
            metavar='<share_server>',
            help=_('ID of share server to start migration.'),
        )
        parser.add_argument(
            'host',
            metavar='<host@backend>',
            help=_(
                "Destination to migrate the share server to. Use "
                "the format '<node_hostname>@<backend_name>'."
            ),
        )
        parser.add_argument(
            '--preserve-snapshots',
            metavar='<True|False>',
            choices=['True', 'False'],
            required=True,
            help=_(
                "Set to True if snapshots must be preserved at "
                "the migration destination."
            ),
        )
        parser.add_argument(
            '--writable',
            metavar='<True|False>',
            choices=['True', 'False'],
            required=True,
            help=_(
                "Enforces migration to keep all its shares writable "
                "while contents are being moved."
            ),
        )
        parser.add_argument(
            '--nondisruptive',
            metavar='<True|False>',
            choices=['True', 'False'],
            required=True,
            help=_("Enforces migration to be nondisruptive."),
        )
        parser.add_argument(
            '--new-share-network',
            metavar='<new_share_network>',
            required=False,
            default=None,
            help=_(
                'Specify a new share network for the share server. Do not '
                'specify this parameter if the migrating share server has '
                'to be retained within its current share network.'
            ),
        )
        parser.add_argument(
            '--check-only',
            action='store_true',
            default=False,
            help=_("Run a dry-run of the share server migration. "),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share
        share_server = osc_utils.find_resource(
            share_client.share_servers, parsed_args.share_server
        )

        if share_client.api_version >= api_versions.APIVersion("2.57"):
            new_share_net_id = None
            result = None
            if parsed_args.new_share_network:
                new_share_net_id = apiutils.find_resource(
                    share_client.share_networks, parsed_args.new_share_network
                ).id
            if parsed_args.check_only:
                result = share_server.migration_check(
                    parsed_args.host,
                    parsed_args.writable,
                    parsed_args.nondisruptive,
                    parsed_args.preserve_snapshots,
                    new_share_net_id,
                )
            if result:
                if parsed_args.formatter == 'table':
                    for k, v in result.items():
                        if isinstance(v, dict):
                            capabilities_list = [v]
                            dict_values = cliutils.convert_dict_list_to_string(
                                capabilities_list
                            )
                            result[k] = dict_values
                return self.dict2columns(result)
            else:
                share_server.migration_start(
                    parsed_args.host,
                    parsed_args.writable,
                    parsed_args.nondisruptive,
                    parsed_args.preserve_snapshots,
                    new_share_net_id,
                )
                return ({}, {})
        else:
            raise exceptions.CommandError(
                "Share Server Migration is only available "
                "with manila API version >= 2.57"
            )



#   Copyright 2021 Red Hat Inc. All rights reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from osc_lib import exceptions as osc_exceptions
from osc_lib import utils as osc_lib_utils

from manilaclient.common.apiclient import exceptions as api_exceptions
from manilaclient.common import cliutils
from manilaclient.osc.v2 import (
    share_snapshot_instances as osc_share_snapshot_instances,
)
from manilaclient.tests.unit.osc import osc_utils
from manilaclient.tests.unit.osc.v2 import fakes as manila_fakes

COLUMNS = ['ID', 'Snapshot ID', 'Status']

COLUMNS_DETAIL = [
    'ID',
    'Snapshot ID',
    'Status',
    'Created At',
    'Updated At',
    'Share ID',
    'Share Instance ID',
    'Progress',
    'Provider Location',
]


class TestShareSnapshotInstance(manila_fakes.TestShare):
    def setUp(self):
        super().setUp()

        self.share_snapshots_mock = (
            self.app.client_manager.share.share_snapshots
        )
        self.share_snapshots_mock.reset_mock()

        self.share_snapshot_instances_mock = (
            self.app.client_manager.share.share_snapshot_instances
        )
        self.share_snapshot_instances_mock.reset_mock()

        self.share_snapshot_instances_el_mock = (
            self.app.client_manager.share.share_snapshot_instance_export_locations  # noqa: E501
        )
        self.share_snapshot_instances_el_mock.reset_mock()


class TestShareSnapshotInstanceList(TestShareSnapshotInstance):
    def setUp(self):
        super().setUp()

        self.share_snapshot_instances = manila_fakes.FakeShareSnapshotIntances.create_share_snapshot_instances(  # noqa: E501
            count=2
        )

        self.share_snapshot_instances_mock.list.return_value = (
            self.share_snapshot_instances
        )

        self.cmd = osc_share_snapshot_instances.ListShareSnapshotInstance(
            self.app, None
        )

    def test_share_snapshot_instance_list(self):
        values = (
            osc_lib_utils.get_dict_properties(s._info, COLUMNS)
            for s in self.share_snapshot_instances
        )

        arglist = []
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.assertEqual(COLUMNS, columns)
        self.assertEqual(list(values), list(data))

    def test_share_snapshot_instance_list_detail(self):
        values = (
            osc_lib_utils.get_dict_properties(s._info, COLUMNS_DETAIL)
            for s in self.share_snapshot_instances
        )

        arglist = ['--detailed']

        verifylist = [('detailed', True)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.assertEqual(COLUMNS_DETAIL, columns)
        self.assertEqual(list(values), list(data))

    def test_share_snapshot_instance_list_snapshot_id(self):
        self.share_snapshot = (
            manila_fakes.FakeShareSnapshot.create_one_snapshot()
        )

        self.share_snapshots_mock.get.return_value = self.share_snapshot
        self.share_snapshot_instances_mock.list.return_value = [
            self.share_snapshot_instances[0]
        ]

        values = (
            osc_lib_utils.get_dict_properties(s._info, COLUMNS)
            for s in [self.share_snapshot_instances[0]]
        )

        arglist = ['--snapshot', self.share_snapshot.id]
        verifylist = [('snapshot', self.share_snapshot.id)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.assertEqual(COLUMNS, columns)
        self.assertEqual(list(values), list(data))


class TestShareSnapshotInstanceShow(TestShareSnapshotInstance):
    def setUp(self):
        super().setUp()

        self.share_snapshot_instance = manila_fakes.FakeShareSnapshotIntances.create_one_snapshot_instance()  # noqa: E501

        self.share_snapshot_instances_mock.get.return_value = (
            self.share_snapshot_instance
        )

        self.share_snapshot_instances_el_list = manila_fakes.FakeShareSnapshotInstancesExportLocations.create_share_snapshot_instances(  # noqa: E501
            count=2
        )

        self.share_snapshot_instances_el_mock.list.return_value = (
            self.share_snapshot_instances_el_list
        )

        self.cmd = osc_share_snapshot_instances.ShowShareSnapshotInstance(
            self.app, None
        )

        self.share_snapshot_instance._info['export_locations'] = (
            cliutils.convert_dict_list_to_string(
                self.share_snapshot_instances_el_list
            )
        )

        self.data = self.share_snapshot_instance._info.values()
        self.columns = self.share_snapshot_instance._info.keys()

    def test_share_snapshot_instance_show_missing_args(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            osc_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_share_snapshot_instance_show(self):
        arglist = [self.share_snapshot_instance.id]
        verifylist = [('snapshot_instance', self.share_snapshot_instance.id)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.share_snapshot_instances_mock.get.assert_called_with(
            self.share_snapshot_instance.id
        )
        self.assertCountEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)


class TestShareSnapshotInstanceSet(TestShareSnapshotInstance):
    def setUp(self):
        super().setUp()

        self.share_snapshot_instance = manila_fakes.FakeShareSnapshotIntances.create_one_snapshot_instance()  # noqa: E501

        self.snapshot_instance_status = 'available'

        self.cmd = osc_share_snapshot_instances.SetShareSnapshotInstance(
            self.app, None
        )

    def test_share_snapshot_instance_set_missing_args(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            osc_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_share_snapshot_instance_set_instance_not_found(self):
        arglist = [
            self.share_snapshot_instance.id,
            '--status',
            self.snapshot_instance_status,
        ]
        verifylist = [
            ('snapshot_instance', self.share_snapshot_instance.id),
            ('status', self.snapshot_instance_status),
        ]

        self.share_snapshot_instances_mock.reset_state.side_effect = (
            api_exceptions.NotFound()
        )

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaises(
            osc_exceptions.CommandError, self.cmd.take_action, parsed_args
        )

    def test_share_snapshot_instance_set(self):
        arglist = [
            self.share_snapshot_instance.id,
            '--status',
            self.snapshot_instance_status,
        ]
        verifylist = [
            ('snapshot_instance', self.share_snapshot_instance.id),
            ('status', self.snapshot_instance_status),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        self.share_snapshot_instances_mock.reset_state.assert_called_with(
            self.share_snapshot_instance.id, self.snapshot_instance_status
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from manilaclient import api_versions
from manilaclient import base
from manilaclient.common import constants

RESOURCES_NAME = 'share_backups'
RESOURCE_NAME = 'share_backup'
RESOURCES_PATH = '/share-backups'
RESOURCE_PATH = '/share-backups/%s'
RESOURCE_PATH_ACTION = '/share-backups/%s/action'


class ShareBackup(base.Resource):
    def __repr__(self):
        return f"<Share Backup: {self.id}>"


class ShareBackupManager(base.ManagerWithFind):
    """Manage :class:`ShareBackup` resources."""

    resource_class = ShareBackup

    @api_versions.wraps("2.80")
    @api_versions.experimental_api
    def get(self, backup):
        """Get a share backup.

        :param backup: either backup object or its UUID.
        :rtype: :class:`ShareBackup`
        """
        backup_id = base.getid(backup)
        return self._get(RESOURCE_PATH % backup_id, RESOURCE_NAME)

    @api_versions.wraps("2.80")
    @api_versions.experimental_api
    def list(
        self, detailed=True, search_opts=None, sort_key=None, sort_dir=None
    ):
        """List all share backups or list backups belonging to a share.

        :param detailed: list backups with detailed fields.
        :param search_opts: Search options to filter out shares.
        :param sort_key: Key to be sorted.
        :param sort_dir: Sort direction, should be 'desc' or 'asc'.
        :rtype: list of :class:`ShareBackup`
        """

        search_opts = search_opts or {}

        if sort_key is not None:
            if sort_key in constants.BACKUP_SORT_KEY_VALUES:
                search_opts['sort_key'] = sort_key
            else:
                raise ValueError(
                    'sort_key must be one of the following: {}.'.format(
                        ', '.join(constants.BACKUP_SORT_KEY_VALUES)
                    )
                )

        if sort_dir is not None:
            if sort_dir in constants.SORT_DIR_VALUES:
                search_opts['sort_dir'] = sort_dir
            else:
                raise ValueError(
                    'sort_dir must be one of the following: {}.'.format(
                        ', '.join(constants.SORT_DIR_VALUES)
                    )
                )

        query_string = self._build_query_string(search_opts)
        if detailed:
            path = f"/share-backups/detail{query_string}"
        else:
            path = f"/share-backups{query_string}"

        return self._list(path, 'share_backups')

    @api_versions.wraps("2.80")
    @api_versions.experimental_api
    def create(self, share, backup_options=None, description=None, name=None):
        """Create a backup for a share.

        :param share: The share to create the backup of. Can be the share
        object or its UUID.
        :param backup_options: dict - custom set of key-values
        :param name: text - name of new share
        :param description: - description for new share
        """
        share_id = base.getid(share)
        body = {
            'share_id': share_id,
            'backup_options': backup_options,
            'description': description,
            'name': name,
        }

        return self._create(
            RESOURCES_PATH, {RESOURCE_NAME: body}, RESOURCE_NAME
        )

    @api_versions.wraps("2.80")
    @api_versions.experimental_api
    def delete(self, backup):
        backup_id = base.getid(backup)
        url = RESOURCE_PATH % backup_id
        self._delete(url)

    @api_versions.wraps("2.80", "2.90")
    @api_versions.experimental_api
    def restore(self, backup_id):
        return self._action('restore', backup_id)

    @api_versions.wraps("2.91")
    @api_versions.experimental_api
    def restore(self, backup_id, target_share_id=None):  # noqa F811
        return self._action('restore', backup_id, info=target_share_id)

    @api_versions.wraps("2.80")
    @api_versions.experimental_api
    def reset_status(self, backup, state):
        return self._action('reset_status', backup, {"status": state})

    @api_versions.wraps("2.80")
    @api_versions.experimental_api
    def update(self, backup, **kwargs):
        if not kwargs:
            return
        backup_id = base.getid(backup)
        body = {'share_backup': kwargs}
        return self._update(RESOURCE_PATH % backup_id, body)

    def _action(self, action, backup, info=None, **kwargs):
        body = {action: info}
        self.run_hooks('modify_body_for_action', body, **kwargs)
        backup_id = base.getid(backup)
        url = RESOURCE_PATH_ACTION % backup_id
        return self.api.client.post(url, body=body)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from debtcollector import removals

from keystoneauth1 import adapter
from keystoneauth1 import identity
from keystoneauth1 import session

import manilaclient
from manilaclient.common import constants
from manilaclient.common import httpclient
from manilaclient import exceptions
from manilaclient.v2 import limits
from manilaclient.v2 import quota_classes
from manilaclient.v2 import quotas
from manilaclient.v2 import scheduler_stats
from manilaclient.v2 import security_services
from manilaclient.v2 import services
from manilaclient.v2 import share_networks
from manilaclient.v2 import share_servers
from manilaclient.v2 import share_snapshots
from manilaclient.v2 import share_type_access
from manilaclient.v2 import share_types
from manilaclient.v2 import shares


@removals.removed_class(
    "Client", message="Please use 'v2.Client' instead", removal_version='2.0.0'
)
class Client:
    """Top-level object to access the OpenStack Manila API.

    Create an instance with your creds::

        >>> client = Client(username=USERNAME,
                            password=PASSWORD,
                            project_name=PROJECT_NAME,
                            auth_url=AUTH_URL)

    Or, alternatively, you can create a client instance using the
    keystoneauth1.session API::

        >>> from keystoneauth1.identity import v3
        >>> from keystoneauth1 import session
        >>> from manilaclient import client
        >>> auth = v3.Password(auth_url=AUTH_URL,
                               username=USERNAME,
                               user_domain_name=USER_DOMAIN_NAME,
                               password=PASSWORD,
                               project_name=PROJECT_ID,
                               project_domain_name=PROJECT_DOMAIN_NAME)
        >>> sess = session.Session(auth=auth)
        >>> manila = client.Client(VERSION, session=sess)

    Then call methods on its managers::

        >>> client.shares.list()
        ...
    """

    @removals.removed_kwarg(
        'use_keyring',
        message='This parameter is no longer supported and has no effect.',
        version='5.8.0',
        removal_version='6.0.0',
    )
    @removals.removed_kwarg(
        'force_new_token',
        message='This parameter is no longer supported and has no effect.',
        version='5.8.0',
        removal_version='6.0.0',
    )
    @removals.removed_kwarg(
        'cached_token_lifetime',
        message='This parameter is no longer supported and has no effect.',
        version='5.8.0',
        removal_version='6.0.0',
    )
    def __init__(
        self,
        username=None,
        project_id=None,
        auth_url=None,
        insecure=False,
        timeout=None,
        tenant_id=None,
        project_name=None,
        region_name=None,
        endpoint_type='publicURL',
        extensions=None,
        service_type=constants.V1_SERVICE_TYPE,
        service_name=None,
        retries=None,
        http_log_debug=False,
        input_auth_token=None,
        session=None,
        auth=None,
        cacert=None,
        service_catalog_url=None,
        user_agent='python-manilaclient',
        use_keyring=False,
        force_new_token=False,
        cached_token_lifetime=300,
        api_version=manilaclient.API_DEPRECATED_VERSION,
        user_id=None,
        user_domain_id=None,
        user_domain_name=None,
        project_domain_id=None,
        project_domain_name=None,
        cert=None,
        password=None,
        **kwargs,
    ):
        self.username = username
        self.password = password
        self.tenant_id = tenant_id or project_id
        self.tenant_name = project_name

        self.user_id = user_id
        self.project_id = project_id or tenant_id
        self.project_name = project_name
        self.user_domain_id = user_domain_id
        self.user_domain_name = user_domain_name
        self.project_domain_id = project_domain_id
        self.project_domain_name = project_domain_name

        self.endpoint_type = endpoint_type
        self.auth_url = auth_url
        self.region_name = region_name

        self.cacert = cacert
        self.cert = cert
        self.insecure = insecure

        if input_auth_token and not service_catalog_url:
            msg = (
                "For token-based authentication you should "
                "provide 'input_auth_token' and 'service_catalog_url'."
            )
            raise exceptions.ClientException(msg)

        self.project_id = tenant_id if tenant_id is not None else project_id
        self.keystone_client = None
        self.session = session

        # NOTE(u_glide): token authorization has highest priority.
        # That's why session and/or password will be ignored
        # if token is provided.
        if not input_auth_token:
            if session:
                # Modern path - session provided by caller (e.g., OSC plugin)
                self.keystone_client = adapter.LegacyJsonAdapter(
                    session=session,
                    auth=auth,
                    interface=endpoint_type,
                    service_type=service_type,
                    service_name=service_name,
                    region_name=region_name,
                )
                input_auth_token = self.keystone_client.session.get_token(auth)
            else:
                # Legacy path - create auth plugin and session ourselves
                auth, ks_session = self._get_keystone_auth_and_session()
                self.keystone_client = adapter.LegacyJsonAdapter(
                    session=ks_session,
                    auth=auth,
                    interface=endpoint_type,
                    service_type=service_type,
                    service_name=service_name,
                    region_name=region_name,
                )
                input_auth_token = self.keystone_client.session.get_token(auth)

        if not input_auth_token:
            raise RuntimeError("Not Authorized")

        if not service_catalog_url:
            # Use keystoneauth1 session endpoint discovery
            service_catalog_url = self.keystone_client.session.get_endpoint(
                self.keystone_client.auth,
                interface=endpoint_type,
                service_type=service_type,
                region_name=region_name,
            )

        if not service_catalog_url:
            raise RuntimeError("Could not find Manila endpoint in catalog")

        self.api_version = api_version
        self.client = httpclient.HTTPClient(
            service_catalog_url,
            input_auth_token,
            user_agent,
            insecure=insecure,
            cacert=cacert,
            cert=cert,
            timeout=timeout,
            retries=retries,
            http_log_debug=http_log_debug,
            api_version=self.api_version,
        )

        self.limits = limits.LimitsManager(self)
        self.services = services.ServiceManager(self)
        self.security_services = security_services.SecurityServiceManager(self)
        self.share_networks = share_networks.ShareNetworkManager(self)

        self.quota_classes = quota_classes.QuotaClassSetManager(self)
        self.quotas = quotas.QuotaSetManager(self)

        self.shares = shares.ShareManager(self)
        self.share_snapshots = share_snapshots.ShareSnapshotManager(self)

        self.share_types = share_types.ShareTypeManager(self)
        self.share_type_access = share_type_access.ShareTypeAccessManager(self)
        self.share_servers = share_servers.ShareServerManager(self)
        self.pools = scheduler_stats.PoolManager(self)

        self._load_extensions(extensions)

    def _load_extensions(self, extensions):
        if not extensions:
            return

        for extension in extensions:
            if extension.manager_class:
                setattr(self, extension.name, extension.manager_class(self))

    def _get_keystone_auth_and_session(self):
        """Create keystoneauth1 auth plugin and session for authentication.

        Returns:
            tuple: (auth_plugin, session) for use with keystoneauth1
        """
        # Create session with SSL settings
        if self.insecure:
            verify = False
        else:
            verify = self.cacert or True
        ks_session = session.Session(verify=verify, cert=self.cert)

        # Discover Keystone v3 endpoint
        ks_discover = session.discover.Discover(ks_session, self.auth_url)
        auth_url = ks_discover.url_for('v3.0')
        if not auth_url:
            raise exceptions.CommandError(
                'Unable to determine the Keystone version to authenticate '
                'with using the given auth_url.'
            )

        # Create v3 Password auth plugin
        auth = identity.v3.Password(
            auth_url=auth_url,
            username=self.username,
            password=self.password,
            user_id=self.user_id,
            user_domain_name=self.user_domain_name,
            user_domain_id=self.user_domain_id,
            project_id=self.project_id or self.tenant_id,
            project_name=self.project_name,
            project_domain_name=self.project_domain_name,
            project_domain_id=self.project_domain_id,
        )

        return auth, ks_session
